import warnings
from typing import Any, Literal, cast

from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.neat._alpha import AlphaFlags
from cognite.neat._client import NeatClient
from cognite.neat._constants import (
    CLASSIC_CDF_NAMESPACE,
    get_default_prefixes_and_namespaces,
)
from cognite.neat._graph import examples as instances_examples
from cognite.neat._graph import extractors
from cognite.neat._graph.transformers import (
    ConvertLiteral,
    LiteralToEntity,
    Transformers,
)
from cognite.neat._graph.transformers._prune_graph import (
    AttachPropertyFromTargetToSource,
    PruneDeadEndEdges,
    PruneInstancesOfUnknownType,
    PruneTypes,
)
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import MissingCogniteClientWarning
from cognite.neat._rules import catalog, importers
from cognite.neat._rules.importers import BaseImporter
from cognite.neat._rules.transformers import ClassicPrepareCore
from cognite.neat._utils.reader import NeatReader

from ._state import SessionState
from ._wizard import NeatObjectType, RDFFileType, XMLFileType, object_wizard, rdf_dm_wizard, xml_format_wizard
from .engine import import_engine
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class ReadAPI:
    """Read from a data source into NeatSession graph store."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.cdf = CDFReadAPI(state, verbose)
        self.rdf = RDFReadAPI(state, verbose)
        self.excel = ExcelReadAPI(state, verbose)
        self.csv = CSVReadAPI(state, verbose)
        self.yaml = YamlReadAPI(state, verbose)
        self.xml = XMLReadAPI(state, verbose)
        self.examples = Examples(state)

    def session(self, io: Any) -> None:
        """Reads a Neat Session from a zip file.

        Args:
            io: file path to the Neat Session

        Example:
            ```python
            neat.read.session("path_to_neat_session")
            ```
        """
        reader = NeatReader.create(io)
        path = reader.materialize_path()

        self._state.instances.store.write(extractors.RdfFileExtractor.from_zip(path))


@session_class_wrapper
class BaseReadAPI:
    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose


@session_class_wrapper
class CDFReadAPI(BaseReadAPI):
    """Reads from CDF Data Models.
    Use the `.data_model()` method to load a CDF Data Model to the knowledge graph.

    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        super().__init__(state, verbose)
        self.classic = CDFClassicAPI(state, verbose)

    @property
    def _get_client(self) -> NeatClient:
        if self._state.client is None:
            raise NeatValueError("No client provided. Please provide a client to read a data model.")
        return self._state.client

    def data_model(self, data_model_id: DataModelIdentifier) -> IssueList:
        """Reads a Data Model from CDF to the knowledge graph.

        Args:
            data_model_id: Tuple of strings with the id of a CDF Data Model.
            Notation as follows (<name_of_space>, <name_of_data_model>, <data_model_version>)

        Example:
            ```python
            neat.read.cdf.data_model(("example_data_model_space", "EXAMPLE_DATA_MODEL", "v1"))
            ```
        """

        data_model_id = DataModelId.load(data_model_id)

        if not data_model_id.version:
            raise NeatSessionError("Data model version is required to read a data model.")

        importer = importers.DMSImporter.from_data_model_id(self._get_client, data_model_id)
        return self._state.rule_import(importer)

    def graph(
        self,
        data_model_id: DataModelIdentifier,
        instance_space: str | SequenceNotStr[str] | None = None,
        skip_cognite_views: bool = True,
    ) -> IssueList:
        """Reads a knowledge graph from Cognite Data Fusion (CDF).

        Args:
            data_model_id: Tuple of strings with the id of a CDF Data Model.
            instance_space: The instance spaces to extract. If None, all instance spaces are extracted.
            skip_cognite_views: If True, all Cognite Views are skipped. For example, if you have the CogniteAsset
                view in you data model, it will ont be used to extract instances.

        Returns:
            IssueList: A list of issues that occurred during the extraction.

        """
        return self._graph(data_model_id, instance_space, skip_cognite_views, unpack_json=False)

    def _graph(
        self,
        data_model_id: DataModelIdentifier,
        instance_space: str | SequenceNotStr[str] | None = None,
        skip_cognite_views: bool = True,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> IssueList:
        extractor = extractors.DMSGraphExtractor.from_data_model_id(
            # We are skipping the Cognite Views
            data_model_id,
            self._get_client,
            instance_space=instance_space,
            skip_cognite_views=skip_cognite_views,
            unpack_json=unpack_json,
            str_to_ideal_type=str_to_ideal_type,
        )
        return self._state.write_graph(extractor)

    def raw(
        self,
        db_name: str,
        table_name: str,
        type: str | None = None,
        foreign_keys: str | SequenceNotStr[str] | None = None,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> IssueList:
        """Reads a raw table from CDF to the knowledge graph.

        Args:
            db_name: The name of the database
            table_name: The name of the table, this will be assumed to be the type of the instances.
            type: The type of instances in the table. If None, the table name will be used.
            foreign_keys: The name of the columns that are foreign keys. If None, no foreign keys are used.
            unpack_json: If True, the JSON objects will be unpacked into the graph.
            str_to_ideal_type: If True, the string values will be converted to ideal types.

        Returns:
            IssueList: A list of issues that occurred during the extraction.

        Example:
            ```python
            neat.read.cdf.raw("my_db", "my_table", "Asset")
            ```

        """
        extractor = extractors.RAWExtractor(
            self._get_client,
            db_name=db_name,
            table_name=table_name,
            table_type=type,
            foreign_keys=foreign_keys,
            unpack_json=unpack_json,
            str_to_ideal_type=str_to_ideal_type,
        )
        return self._state.instances.store.write(extractor)


@session_class_wrapper
class CDFClassicAPI(BaseReadAPI):
    """Reads from the Classic Data Model from CDF.
    Use the `.graph()` method to load CDF core resources to the knowledge graph.

    """

    @property
    def _get_client(self) -> NeatClient:
        if self._state.client is None:
            raise ValueError("No client provided. Please provide a client to read a data model.")
        return self._state.client

    def graph(
        self,
        root_asset_external_id: str,
        limit_per_type: int | None = None,
        identifier: Literal["id", "externalId"] = "id",
    ) -> IssueList:
        """Reads the classic knowledge graph from CDF.

        The Classic Graph consists of the following core resource type.

        !!! note "Classic Node CDF Resources"
             - Assets
             - TimeSeries
             - Sequences
             - Events
             - Files

        All the classic node CDF resources can have one or more connections to one or more assets. This
        will match a direct relationship in the data modeling of CDF.

        In addition, you have relationships between the classic node CDF resources. This matches an edge
        in the data modeling of CDF.

        Finally, you have labels and data sets that to organize the graph. In which data sets have a similar,
        but different, role as a space in data modeling. While labels can be compared to node types in data modeling,
        used to quickly filter and find nodes/edges.

        This extractor will extract the classic CDF graph into Neat starting from either a data set or a root asset.

        It works as follows:
            1. Extract all core nodes (assets, time series, sequences, events, files) filtered by the given data set or
               root asset.
            2. Extract all relationships starting from any of the extracted core nodes.
            3. Extract all core nodes that are targets of the relationships that are not already extracted.
            4. Extract all labels that are connected to the extracted core nodes/relationships.
            5. Extract all data sets that are connected to the extracted core nodes/relationships.

        Args:
            root_asset_external_id: The external id of the root asset
            limit_per_type: The maximum number of nodes to extract per core node type. If None, all nodes are extracted.
            identifier: The identifier to use for the core nodes. Note selecting "id" can cause issues if the external
                ID of the core nodes is missing. Default is "id".

        Returns:
            IssueList: A list of issues that occurred during the extraction.

        Example:
            ```python
            neat.read.cdf.graph("root_asset_external_id")
            ```
        """
        return self._graph(
            root_asset_external_id, limit_per_type, identifier, reference_timeseries=False, reference_files=False
        )

    def _graph(
        self,
        root_asset_external_id: str,
        limit_per_type: int | None = None,
        identifier: Literal["id", "externalId"] = "id",
        reference_timeseries: bool = False,
        reference_files: bool = False,
        unpack_metadata: bool = False,
        skip_sequence_rows: bool = False,
    ) -> IssueList:
        namespace = CLASSIC_CDF_NAMESPACE
        extractor = extractors.ClassicGraphExtractor(
            self._get_client,
            root_asset_external_id=root_asset_external_id,
            limit_per_type=limit_per_type,
            namespace=namespace,
            prefix="Classic",
            identifier=identifier,
            unpack_metadata=unpack_metadata,
            skip_sequence_rows=skip_sequence_rows,
        )
        self._state.instances.neat_prefix_by_predicate_uri.update(extractor.neat_prefix_by_predicate_uri)
        self._state.instances.neat_prefix_by_type_uri.update(extractor.neat_prefix_by_type_uri)
        extract_issues = self._state.write_graph(extractor)
        if identifier == "externalId":
            self._state.quoted_source_identifiers = True

        self._state.instances.store.transform(
            ConvertLiteral(
                namespace["ClassicTimeSeries"],
                namespace["isString"],
                lambda is_string: "string" if is_string else "numeric",
            )
        )
        self._state.instances.store.transform(
            LiteralToEntity(None, namespace["source"], "ClassicSourceSystem", "name"),
        )
        # Updating the information model.
        prepare_issues = self._state.rule_store.transform(
            ClassicPrepareCore(namespace, reference_timeseries, reference_files)
        )

        all_issues = IssueList(extract_issues + prepare_issues)
        # Update the provenance with all issue.
        object.__setattr__(self._state.instances.store.provenance[-1].target_entity, "issues", all_issues)
        all_issues.action = "Read Classic Graph"
        if all_issues:
            print("Use the .inspect.issues() for more details")

        return all_issues


@session_class_wrapper
class ExcelReadAPI(BaseReadAPI):
    """Reads a Neat Excel Rules sheet to the graph store. The rules sheet may stem from an Information architect,
    or a DMS Architect.

    Args:
        io: file path to the Excel sheet

    Example:
        ```python
        neat.read.excel("information_or_dms_rules_sheet.xlsx")
        ```
    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        super().__init__(state, verbose)

    def __call__(self, io: Any, enable_manual_edit: bool = False) -> IssueList:
        """Reads a Neat Excel Rules sheet to the graph store. The rules sheet may stem from an Information architect,
        or a DMS Architect.

        Args:
            io: file path to the Excel sheet
            enable_manual_edit: If True, the user will be able to re-import rules which where edit outside NeatSession

        !!! note "Manual Edit Warning"
            This is an alpha feature and is subject to change without notice.
            It is expected to have some limitations and may not work as expected in all cases.
        """
        reader = NeatReader.create(io)
        path = reader.materialize_path()

        if enable_manual_edit:
            warnings.filterwarnings("default")
            AlphaFlags.manual_rules_edit.warn()

        return self._state.rule_import(importers.ExcelImporter(path), enable_manual_edit)


@session_class_wrapper
class YamlReadAPI(BaseReadAPI):
    def __call__(self, io: Any, format: Literal["neat", "toolkit"] = "neat") -> IssueList:
        """Reads a yaml with either neat rules, or several toolkit yaml files to import Data Model(s) into NeatSession.

        Args:
            io: File path to the Yaml file in the case of "neat" yaml, or path to a zip folder or directory with several
                Yaml files in the case of "toolkit".
            format: The format of the yaml file(s). Can be either "neat" or "toolkit".

        Example:
            ```python
            neat.read.yaml("path_to_toolkit_yamls")
            ```
        """
        reader = NeatReader.create(io)
        path = reader.materialize_path()
        importer: BaseImporter
        if format == "neat":
            importer = importers.YAMLImporter.from_file(path, source_name=f"{reader!s}")
        elif format == "toolkit":
            dms_importer = importers.DMSImporter.from_path(path, self._state.client)
            if dms_importer.issue_list.has_warning_type(MissingCogniteClientWarning):
                raise NeatSessionError(
                    "No client provided. You are referencing Cognite containers in your data model, "
                    "NEAT needs a client to lookup the container definitions. "
                    "Please set the client in the session, NeatSession(client=client)."
                )
            importer = dms_importer
        else:
            raise NeatValueError(f"Unsupported YAML format: {format}")
        return self._state.rule_import(importer)


@session_class_wrapper
class CSVReadAPI(BaseReadAPI):
    """Reads a csv that contains a column to use as primary key which will be the unique identifier for the type of
    data you want to read in. Ex. a csv can hold information about assets, and their identifiers are specified in
    a "ASSET_TAG" column.

    Args:
        io: file path or url to the csv
        type: string that specifies what type of data the csv contains. For instance "Asset" or "Equipment"
        primary_key: string name of the column that should be used as the unique identifier for each row of data

    Example:
        ```python
        type_described_in_table = "Turbine"
        column_with_identifier = "UNIQUE_TAG_NAME"
        neat.read.csv("url_or_path_to_csv_file", type=type_described_in_table, primary_key=column_with_identifier)
        ```
    """

    def __call__(self, io: Any, type: str, primary_key: str) -> None:
        warnings.filterwarnings("default")
        AlphaFlags.csv_read.warn()

        engine = import_engine()
        engine.set.format = "csv"
        engine.set.file = NeatReader.create(io).materialize_path()
        engine.set.type = type
        engine.set.primary_key = primary_key
        extractor = engine.create_extractor()

        self._state.instances.store.write(extractor)


@session_class_wrapper
class XMLReadAPI(BaseReadAPI):
    """Reads an XML file that is either of DEXPI or AML format.

    Args:
        io: file path or url to the XML
        format: can be either "dexpi" or "aml" are the currenly supported XML source types.
    """

    def __call__(
        self,
        io: Any,
        format: XMLFileType | None = None,
    ) -> None:
        path = NeatReader.create(io).materialize_path()
        if format is None:
            format = xml_format_wizard()

        if format.lower() == "dexpi":
            return self.dexpi(path)

        if format.lower() == "aml":
            return self.aml(path)

        else:
            raise NeatValueError("Only support XML files of DEXPI format at the moment.")

    def dexpi(self, io: Any) -> None:
        """Reads a DEXPI file into the NeatSession and executes set of predefined transformations.

        Args:
            io: file path or url to the DEXPI file

        Example:
            ```python
            neat.read.xml.dexpi("url_or_path_to_dexpi_file")
            ```

        !!! note "This method bundles several graph transformers which"
            - attach values of generic attributes to nodes
            - create associations between nodes
            - remove unused generic attributes
            - remove associations between nodes that do not exist in the extracted graph
            - remove edges to nodes that do not exist in the extracted graph
        """
        warnings.filterwarnings("default")
        AlphaFlags.dexpi_read.warn()

        path = NeatReader.create(io).materialize_path()
        engine = import_engine()
        engine.set.format = "dexpi"
        engine.set.file = path
        extractor = engine.create_extractor()
        self._state.instances.store.write(extractor)

        DEXPI = get_default_prefixes_and_namespaces()["dexpi"]

        transformers = [
            # Remove any instance which type is unknown
            PruneInstancesOfUnknownType(),
            # Directly connect generic attributes
            AttachPropertyFromTargetToSource(
                target_property=DEXPI.Value,
                target_property_holding_new_property=DEXPI.Name,
                target_node_type=DEXPI.GenericAttribute,
                delete_target_node=True,
            ),
            # Directly connect associations
            AttachPropertyFromTargetToSource(
                target_property=DEXPI.ItemID,
                target_property_holding_new_property=DEXPI.Type,
                target_node_type=DEXPI.Association,
                delete_target_node=True,
            ),
            # Remove unused generic attributes and associations
            PruneTypes([DEXPI.GenericAttribute, DEXPI.Association]),
            # Remove edges to nodes that do not exist in the extracted graph
            PruneDeadEndEdges(),
        ]

        for transformer in transformers:
            self._state.instances.store.transform(cast(Transformers, transformer))

    def aml(self, io: Any):
        """Reads an AML file into NeatSession and executes a set of predefined transformations.

        Args:
            io: file path or url to the AML file

        Example:
            ```python
            neat.read.xml.aml("url_or_path_to_aml_file")
            ```

        !!! note "This method bundles several graph transformers which"
            - attach values of attributes to nodes
            - remove unused attributes
            - remove edges to nodes that do not exist in the extracted graph
        """
        warnings.filterwarnings("default")
        AlphaFlags.aml_read.warn()

        path = NeatReader.create(io).materialize_path()
        engine = import_engine()
        engine.set.format = "aml"
        engine.set.file = path
        extractor = engine.create_extractor()
        self._state.instances.store.write(extractor)

        AML = get_default_prefixes_and_namespaces()["aml"]

        transformers = [
            # Remove any instance which type is unknown
            PruneInstancesOfUnknownType(),
            # Directly connect generic attributes
            AttachPropertyFromTargetToSource(
                target_property=AML.Value,
                target_property_holding_new_property=AML.Name,
                target_node_type=AML.Attribute,
                delete_target_node=True,
            ),
            # Prune unused attributes
            PruneTypes([AML.Attribute]),
            # # Remove edges to nodes that do not exist in the extracted graph
            PruneDeadEndEdges(),
        ]

        for transformer in transformers:
            self._state.instances.store.transform(cast(Transformers, transformer))


@session_class_wrapper
class RDFReadAPI(BaseReadAPI):
    """Reads an RDF source into NeatSession. Supported sources are "ontology" or "imf".

    Args:
        io: file path or url to the RDF source
    """

    def __init__(self, state: SessionState, verbose: bool) -> None:
        super().__init__(state, verbose)

    def ontology(self, io: Any) -> IssueList:
        """Reads an OWL ontology source into NeatSession.

        Args:
            io: file path or url to the OWL file

        Example:
            ```python
            neat.read.rdf.ontology("url_or_path_to_owl_source")
            ```
        """
        warnings.filterwarnings("default")
        AlphaFlags.ontology_read.warn()

        reader = NeatReader.create(io)
        importer = importers.OWLImporter.from_file(reader.materialize_path(), source_name=f"file {reader!s}")
        return self._state.rule_import(importer)

    def imf(self, io: Any) -> IssueList:
        """Reads IMF Types provided as SHACL shapes into NeatSession.

        Args:
            io: file path or url to the IMF file

        Example:
            ```python
            neat.read.rdf.imf("url_or_path_to_imf_source")
            ```
        """
        warnings.filterwarnings("default")
        AlphaFlags.imf_read.warn()

        reader = NeatReader.create(io)
        importer = importers.IMFImporter.from_file(reader.materialize_path(), source_name=f"file {reader!s}")
        return self._state.rule_import(importer)

    def instances(self, io: Any) -> IssueList:
        reader = NeatReader.create(io)
        self._state.instances.store.write(extractors.RdfFileExtractor(reader.materialize_path()))
        return IssueList()

    def __call__(
        self,
        io: Any,
        type: NeatObjectType | None = None,
        source: RDFFileType | None = None,
    ) -> IssueList:
        if type is None:
            type = object_wizard()

        type = type.lower()

        if type == "data model":
            source = source or rdf_dm_wizard("What type of data model is the RDF?")
            source = cast(str, source).lower()  # type: ignore

            if source == "ontology":
                return self.ontology(io)
            elif source == "imf types":
                return self.imf(io)
            else:
                raise ValueError(f"Expected ontology, imf types or instances, got {source}")

        elif type == "instances":
            return self.instances(io)

        else:
            raise NeatSessionError(f"Expected data model or instances, got {type}")


@session_class_wrapper
class Examples:
    """Used as example for reading various sources into NeatSession."""

    def __init__(self, state: SessionState) -> None:
        self._state = state

    @property
    def _get_client(self) -> NeatClient:
        if self._state.client is None:
            raise NeatValueError("No client provided. Please provide a client to read a data model.")
        return self._state.client

    def nordic44(self) -> IssueList:
        """Reads the Nordic 44 knowledge graph into the NeatSession graph store."""
        self._state.instances.store.write(extractors.RdfFileExtractor(instances_examples.nordic44_knowledge_graph))
        return IssueList()

    def pump_example(self) -> IssueList:
        """Reads the Hello World pump example into the NeatSession."""
        importer: importers.ExcelImporter = importers.ExcelImporter(catalog.hello_world_pump)
        return self._state.rule_import(importer)

    def core_data_model(self) -> IssueList:
        """Reads the core data model example into the NeatSession."""

        cdm_v1 = DataModelId.load(("cdf_cdm", "CogniteCore", "v1"))
        importer: importers.DMSImporter = importers.DMSImporter.from_data_model_id(self._get_client, cdm_v1)
        return self._state.rule_import(importer)



import inspect
import sys
from abc import ABC, abstractmethod
from functools import total_ordering
from types import UnionType
from typing import Any, ClassVar, Generic, Literal, TypeVar, Union, cast, get_args, get_origin, overload

from cognite.client.data_classes.data_modeling import DirectRelationReference
from cognite.client.data_classes.data_modeling.data_types import UnitReference
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    DataModelId,
    NodeId,
    PropertyId,
    ViewId,
)
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    model_serializer,
    model_validator,
)

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._utils.text import replace_non_alphanumeric_with_underscore

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self

from cognite.neat._rules._constants import (
    ENTITY_PATTERN,
    SPLIT_ON_COMMA_PATTERN,
    SPLIT_ON_EDGE_ENTITY_ARGS_PATTERN,
    SPLIT_ON_EQUAL_PATTERN,
    EntityTypes,
)

from ._constants import (
    _PARSE,
    Undefined,
    Unknown,
    _UndefinedType,
    _UnknownType,
)


@total_ordering
class Entity(BaseModel, extra="ignore"):
    """Entity is a class or property in OWL/RDF sense."""

    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: str | _UndefinedType = Undefined
    suffix: str

    @classmethod
    @overload
    def load(cls: "type[T_Entity]", data: Any, strict: Literal[True], **defaults) -> "T_Entity": ...

    @classmethod
    @overload
    def load(
        cls: "type[T_Entity]", data: Any, strict: Literal[False] = False, **defaults
    ) -> "T_Entity | UnknownEntity": ...

    @classmethod
    def load(cls: "type[T_Entity]", data: Any, strict: bool = False, **defaults) -> "T_Entity | UnknownEntity":
        if isinstance(data, cls):
            return data
        elif isinstance(data, str) and data == str(Unknown):
            if strict:
                raise NeatValueError(f"Failed to load entity {data!s}")
            return UnknownEntity(prefix=Undefined, suffix=Unknown)
        if defaults and isinstance(defaults, dict):
            # This is a trick to pass in default values
            try:
                return cls.model_validate({_PARSE: data, "defaults": defaults})
            except ValueError:
                raise
        else:
            return cls.model_validate(data)

    @model_validator(mode="before")
    def _load(cls, data: Any) -> "dict | Entity":
        defaults = {}
        if isinstance(data, dict) and _PARSE in data:
            defaults = data.get("defaults", {})
            data = data[_PARSE]
        if isinstance(data, dict):
            data.update(defaults)
            return data
        elif hasattr(data, "versioned_id"):
            # Todo: Remove. Is here for backwards compatibility
            data = data.versioned_id
        elif not isinstance(data, str):
            raise ValueError(f"Cannot load {cls.__name__} from {data}")
        elif data == str(Unknown) and cls.type_ == EntityTypes.undefined:
            return dict(prefix=Undefined, suffix=Unknown)  # type: ignore[arg-type]
        elif data == str(Unknown):
            raise ValueError(f"Unknown is not allowed for {cls.type_} entity")

        result = cls._parse(data, defaults)
        output = defaults.copy()
        # Populate by alias
        for field_name, field_ in cls.model_fields.items():
            name = field_.alias or field_name
            if (field_value := result.get(field_name)) and not (field_value in [Unknown, Undefined] and name in output):
                output[name] = result.pop(field_name)
            elif name not in output and name in result:
                output[name] = result.pop(name)
        return output

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, list):
            return [entry.strip() if isinstance(entry, str) else entry for entry in value]
        return value

    @classmethod
    def _parse(cls, raw: str, defaults: dict) -> dict:
        if not (result := ENTITY_PATTERN.match(raw)):
            return dict(prefix=Undefined, suffix=Unknown)
        prefix = result.group("prefix") or Undefined
        suffix = result.group("suffix")
        content = result.group("content")
        if content is None:
            return dict(prefix=prefix, suffix=suffix)
        try:
            if cls == EdgeEntity:
                matches = SPLIT_ON_EDGE_ENTITY_ARGS_PATTERN.findall(content)
                extra_args = {key: value for key, value in matches}
            else:
                extra_args = dict(
                    SPLIT_ON_EQUAL_PATTERN.split(pair.strip()) for pair in SPLIT_ON_COMMA_PATTERN.split(content)
                )
        except ValueError:
            raise NeatValueError(f"Invalid {cls.type_.value} entity: {raw!r}") from None
        expected_args = {
            field_.alias or field_name: field_.annotation for field_name, field_ in cls.model_fields.items()
        }
        for key in list(extra_args):
            if key not in expected_args:
                # Todo Warning about unknown key
                del extra_args[key]
                continue
            annotation = expected_args[key]
            if isinstance(annotation, UnionType) or get_origin(annotation) is Union:
                annotation = get_args(annotation)[0]

            if inspect.isclass(annotation) and issubclass(annotation, Entity):  # type: ignore[arg-type]
                extra_args[key] = annotation.load(extra_args[key], **defaults)  # type: ignore[union-attr, assignment]
        return dict(prefix=prefix, suffix=suffix, **extra_args)

    def dump(self, **defaults: Any) -> str:
        return self._as_str(**defaults)

    def as_tuple(self) -> tuple[str, ...]:
        # We haver overwritten the serialization to str, so we need to do it manually
        extra: tuple[str, ...] = tuple(
            [
                str(v or "")
                for field_name in self.model_fields
                if isinstance(v := getattr(self, field_name), str | None) and field_name not in {"prefix", "suffix"}
            ]
        )
        if isinstance(self.prefix, _UndefinedType):
            return str(self.suffix), *extra
        else:
            return self.prefix, str(self.suffix), *extra

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Entity):
            return NotImplemented
        return self.as_tuple() < other.as_tuple()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Entity):
            return NotImplemented
        return self.as_tuple() == other.as_tuple()

    def __hash__(self) -> int:
        return hash(str(self))

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        # We have overwritten the serialization to str, so we need to do it manually
        model_dump = ((k, v) for k in self.model_fields if (v := getattr(self, k)) is not None)
        args = ",".join([f"{k}={v}" for k, v in model_dump])
        return f"{self.type_.value}({args})"

    @property
    def id(self) -> str:
        return self._as_str()

    def _as_str(self, **defaults: Any) -> str:
        # We have overwritten the serialization to str, so we need to do it manually
        model_dump = {
            field.alias or field_name: v.dump(**defaults) if isinstance(v, Entity) else v
            for field_name, field in self.model_fields.items()
            if (v := getattr(self, field_name)) is not None and field_name not in {"prefix", "suffix"}
        }
        # We only remove the default values if all the fields are default
        # For example, if we dump `cdf_cdm:CogniteAsset(version=v1)` and the default is `version=v1`,
        # we should not remove it unless the space is `cdf_cdm`
        to_delete: list[str] = []
        is_removing_defaults = True
        if isinstance(defaults, dict):
            for key, value in defaults.items():
                if key not in model_dump:
                    continue
                if model_dump[key] == value:
                    to_delete.append(key)
                elif isinstance(self, EdgeEntity):
                    # Exception is edge entity, then, we remove all the defaults we can.
                    continue
                else:
                    # Not all fields are default. We should not remove any of them.
                    is_removing_defaults = False
                    break
        if isinstance(defaults, dict) and self.prefix == defaults.get("prefix") and is_removing_defaults:
            base_id = str(self.suffix)
        elif self.prefix == Undefined:
            base_id = str(self.suffix)
        else:
            is_removing_defaults = False
            base_id = f"{self.prefix}:{self.suffix!s}"
        if is_removing_defaults:
            for key in to_delete:
                del model_dump[key]
        # Sorting to ensure deterministic order
        args = ",".join(f"{k}={v}" for k, v in sorted(model_dump.items(), key=lambda x: x[0]))
        if args:
            return f"{base_id}({args})"
        else:
            return base_id

    @property
    def versioned_id(self) -> str:
        # Todo: Remove. Is here for backwards compatibility
        return self.id

    def as_non_versioned_entity(self) -> str:
        # Todo: Remove. Is here for backwards compatibility
        if self.prefix is Undefined:
            return f"{self.suffix!s}"
        return f"{self.prefix}:{self.suffix!s}"

    def as_dms_compliant_entity(self) -> "Self":
        new_entity = self.model_copy(deep=True)
        new_entity.suffix = replace_non_alphanumeric_with_underscore(new_entity.suffix)
        return new_entity


T_Entity = TypeVar("T_Entity", bound=Entity)


class ClassEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.class_
    version: str | None = None

    def as_view_entity(self, default_space: str, default_version) -> "ViewEntity":
        if self.version is None:
            version = default_version
        else:
            version = self.version
        space = default_space if isinstance(self.prefix, _UndefinedType) else self.prefix
        return ViewEntity(space=space, externalId=str(self.suffix), version=version)

    def as_container_entity(self, default_space: str) -> "ContainerEntity":
        space = default_space if isinstance(self.prefix, _UndefinedType) else self.prefix
        return ContainerEntity(space=space, externalId=str(self.suffix))


class UnknownEntity(ClassEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: _UndefinedType = Undefined
    suffix: _UnknownType = Unknown  # type: ignore[assignment]

    @property
    def id(self) -> str:
        return str(Unknown)


class UnitEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.unit
    prefix: str
    suffix: str

    def as_reference(self) -> UnitReference:
        return UnitReference(external_id=f"{self.prefix}:{self.suffix}")


class AssetFields(StrEnum):
    externalId = "externalId"
    name = "name"
    parentExternalId = "parentExternalId"
    description = "description"
    metadata = "metadata"


class AssetEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.asset
    suffix: str = "Asset"
    prefix: _UndefinedType = Undefined
    property_: AssetFields = Field(alias="property")


class RelationshipEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.relationship
    suffix: str = "Relationship"
    prefix: _UndefinedType = Undefined
    label: str | None = None


T_ID = TypeVar("T_ID", bound=ContainerId | ViewId | DataModelId | PropertyId | NodeId | None)


class DMSEntity(Entity, Generic[T_ID], ABC):
    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: str = Field(alias="space")
    suffix: str = Field(alias="externalId")

    def dump(self, **defaults: Any) -> str:
        if isinstance(defaults, dict) and "space" in defaults:
            defaults["prefix"] = defaults.pop("space")
        return super().dump(**defaults)

    @classmethod  # type: ignore[override]
    @overload
    def load(cls: "type[T_DMSEntity]", data: Any, strict: Literal[True], **defaults) -> "T_DMSEntity": ...

    @classmethod
    @overload
    def load(
        cls: "type[T_DMSEntity]", data: Any, strict: Literal[False] = False, **defaults
    ) -> "T_DMSEntity | DMSUnknownEntity": ...

    @classmethod
    def load(cls: "type[T_DMSEntity]", data: Any, strict: bool = False, **defaults) -> "T_DMSEntity | DMSUnknownEntity":  # type: ignore[override]
        if isinstance(data, str) and data == str(Unknown):
            if strict:
                raise NeatValueError(f"Failed to load entity {data!s}")
            return DMSUnknownEntity.from_id(None)
        return cast(T_DMSEntity, super().load(data, **defaults))

    @property
    def space(self) -> str:
        """Returns entity space in CDF."""
        return self.prefix

    @property
    def external_id(self) -> str:
        """Returns entity external id in CDF."""
        return self.suffix

    @abstractmethod
    def as_id(self) -> T_ID:
        raise NotImplementedError("Method as_id must be implemented in subclasses")

    @classmethod
    @abstractmethod
    def from_id(cls, id: T_ID) -> Self:
        raise NotImplementedError("Method from_id must be implemented in subclasses")

    def as_class(self) -> ClassEntity:
        return ClassEntity(prefix=self.space, suffix=self.external_id)

    def as_dms_compliant_entity(self) -> "Self":
        new_entity = self.model_copy(deep=True)
        new_entity.suffix = replace_non_alphanumeric_with_underscore(new_entity.suffix)
        return new_entity


T_DMSEntity = TypeVar("T_DMSEntity", bound=DMSEntity)


class ContainerEntity(DMSEntity[ContainerId]):
    type_: ClassVar[EntityTypes] = EntityTypes.container

    def as_id(self) -> ContainerId:
        return ContainerId(space=self.space, external_id=self.external_id)

    @classmethod
    def from_id(cls, id: ContainerId) -> "ContainerEntity":
        return cls(space=id.space, externalId=id.external_id)


class DMSVersionedEntity(DMSEntity[T_ID], ABC):
    version: str

    def as_class(self, skip_version: bool = False) -> ClassEntity:
        if skip_version:
            return ClassEntity(prefix=self.space, suffix=self.external_id)
        return ClassEntity(prefix=self.space, suffix=self.external_id, version=self.version)


class ViewEntity(DMSVersionedEntity[ViewId]):
    type_: ClassVar[EntityTypes] = EntityTypes.view

    def as_id(
        self,
    ) -> ViewId:
        return ViewId(space=self.space, external_id=self.external_id, version=self.version)

    def to_property_id(self, property_id: str) -> PropertyId:
        return PropertyId(source=self.as_id(), property=property_id)

    @classmethod
    def from_id(cls, id: ViewId, default_version: str | None = None) -> "ViewEntity":
        if id.version is not None:
            return cls(space=id.space, externalId=id.external_id, version=id.version)
        elif default_version is not None:
            return cls(space=id.space, externalId=id.external_id, version=default_version)
        else:
            raise ValueError("Version must be specified")


class DMSUnknownEntity(DMSEntity[None]):
    """This is a special entity that represents an unknown entity.

    The use case is for direct relations where the source is not known."""

    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: _UndefinedType = Field(Undefined, alias="space")  # type: ignore[assignment]
    suffix: _UnknownType = Field(Unknown, alias="externalId")  # type: ignore[assignment]

    def as_id(self) -> None:
        return None

    @classmethod
    def from_id(cls, id: None) -> "DMSUnknownEntity":
        return cls(space=Undefined, externalId=Unknown)

    @property
    def id(self) -> str:
        return str(Unknown)


class DataModelEntity(DMSVersionedEntity[DataModelId]):
    type_: ClassVar[EntityTypes] = EntityTypes.datamodel

    def as_id(self) -> DataModelId:
        return DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    @classmethod
    def from_id(cls, id: DataModelId) -> "DataModelEntity":
        if id.version is None:
            raise ValueError("Version must be specified")
        return cls(space=id.space, externalId=id.external_id, version=id.version)


class DMSNodeEntity(DMSEntity[NodeId]):
    type_: ClassVar[EntityTypes] = EntityTypes.dms_node

    def as_id(self) -> NodeId:
        return NodeId(space=self.space, external_id=self.external_id)

    def as_reference(self) -> DirectRelationReference:
        return DirectRelationReference(space=self.space, external_id=self.external_id)

    @classmethod
    def from_id(cls, id: NodeId) -> "DMSNodeEntity":
        return cls(space=id.space, externalId=id.external_id)

    @classmethod
    def from_reference(cls, ref: DirectRelationReference) -> "DMSNodeEntity":
        return cls(space=ref.space, externalId=ref.external_id)


class EdgeEntity(DMSEntity[None]):
    type_: ClassVar[EntityTypes] = EntityTypes.edge
    prefix: _UndefinedType = Undefined  # type: ignore[assignment]
    suffix: Literal["edge"] = "edge"
    edge_type: DMSNodeEntity | None = Field(None, alias="type")
    properties: ViewEntity | None = None
    direction: Literal["outwards", "inwards"] = "outwards"

    def dump(self, **defaults: Any) -> str:
        # Add default direction
        return super().dump(**defaults, direction="outwards")

    def as_id(self) -> None:
        return None

    @classmethod
    def from_id(cls, id: None) -> Self:
        return cls()


class ReverseConnectionEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.reverse
    prefix: _UndefinedType = Undefined
    suffix: Literal["reverse"] = "reverse"
    property_: str = Field(alias="property")


class ReferenceEntity(ClassEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.reference_entity
    prefix: str
    property_: str | None = Field(None, alias="property")

    @classmethod
    def from_entity(cls, entity: Entity, property_: str) -> "ReferenceEntity":
        if isinstance(entity, ClassEntity):
            return cls(
                prefix=str(entity.prefix),
                suffix=entity.suffix,
                version=entity.version,
                property=property_,
            )
        else:
            return cls(prefix=str(entity.prefix), suffix=entity.suffix, property=property_)

    def as_view_id(self) -> ViewId:
        if isinstance(self.prefix, _UndefinedType) or isinstance(self.suffix, _UnknownType):
            raise ValueError("Prefix is not defined or suffix is unknown")
        return ViewId(space=self.prefix, external_id=self.suffix, version=self.version)

    def as_view_property_id(self) -> PropertyId:
        if self.property_ is None or self.prefix is Undefined or self.suffix is Unknown:
            raise ValueError("Property is not defined or prefix is not defined or suffix is unknown")
        return PropertyId(source=self.as_view_id(), property=self.property_)

    def as_node_id(self) -> NodeId:
        return NodeId(space=self.prefix, external_id=self.suffix)

    def as_node_entity(self) -> DMSNodeEntity:
        return DMSNodeEntity(space=self.prefix, externalId=self.suffix)

    def as_class_entity(self) -> ClassEntity:
        return ClassEntity(prefix=self.prefix, suffix=self.suffix, version=self.version)



from abc import ABC

from cognite.neat._client import NeatClient
from cognite.neat._issues import MultiValueError, catch_issues
from cognite.neat._issues.errors import NeatTypeError, NeatValueError
from cognite.neat._rules._shared import (
    ReadRules,
    T_ReadInputRules,
    T_VerifiedRules,
    VerifiedRules,
)
from cognite.neat._rules.models import (
    DMSInputRules,
    DMSRules,
    InformationInputRules,
    InformationRules,
)
from cognite.neat._rules.models.dms import DMSValidation
from cognite.neat._rules.models.information import InformationValidation

from ._base import RulesTransformer


class VerificationTransformer(RulesTransformer[T_ReadInputRules, T_VerifiedRules], ABC):
    """Base class for all verification transformers."""

    _rules_cls: type[T_VerifiedRules]
    _validation_cls: type

    def __init__(self, validate: bool = True, client: NeatClient | None = None) -> None:
        self.validate = validate
        self._client = client

    def transform(self, rules: T_ReadInputRules) -> T_VerifiedRules:
        in_ = rules.rules
        if in_ is None:
            raise NeatValueError("Cannot verify rules. The reading of the rules failed.")
        verified_rules: T_VerifiedRules | None = None
        # We need to catch issues as we use the error args to provide extra context for the errors/warnings
        # For example, which row in the spreadsheet the error occurred.
        with catch_issues(rules.read_context) as issues:
            rules_cls = self._get_rules_cls(rules)
            dumped = in_.dump()
            verified_rules = rules_cls.model_validate(dumped)  # type: ignore[assignment]
            if self.validate:
                validation_cls = self._get_validation_cls(verified_rules)  # type: ignore[arg-type]
                if issubclass(validation_cls, DMSValidation):
                    validation_issues = DMSValidation(verified_rules, self._client, rules.read_context).validate()  # type: ignore[arg-type]
                elif issubclass(validation_cls, InformationValidation):
                    validation_issues = InformationValidation(verified_rules, rules.read_context).validate()  # type: ignore[arg-type]
                else:
                    raise NeatValueError("Unsupported rule type")
                issues.extend(validation_issues)

        # Raise issues which is expected to be handled outside of this method
        issues.trigger_warnings()
        if issues.has_errors:
            raise MultiValueError(issues.errors)
        if verified_rules is None:
            raise NeatValueError("Rules were not verified")
        return verified_rules

    def _get_rules_cls(self, in_: T_ReadInputRules) -> type[T_VerifiedRules]:
        return self._rules_cls

    def _get_validation_cls(self, rules: T_VerifiedRules) -> type:
        return self._validation_cls

    @property
    def description(self) -> str:
        return "Verify rules"


class VerifyDMSRules(VerificationTransformer[ReadRules[DMSInputRules], DMSRules]):
    """Class to verify DMS rules."""

    _rules_cls = DMSRules
    _validation_cls = DMSValidation

    def transform(self, rules: ReadRules[DMSInputRules]) -> DMSRules:
        return super().transform(rules)


class VerifyInformationRules(VerificationTransformer[ReadRules[InformationInputRules], InformationRules]):
    """Class to verify Information rules."""

    _rules_cls = InformationRules
    _validation_cls = InformationValidation

    def transform(self, rules: ReadRules[InformationInputRules]) -> InformationRules:
        return super().transform(rules)


class VerifyAnyRules(VerificationTransformer[T_ReadInputRules, VerifiedRules]):
    """Class to verify arbitrary rules"""

    def _get_rules_cls(self, in_: T_ReadInputRules) -> type[VerifiedRules]:
        if isinstance(in_.rules, InformationInputRules):
            return InformationRules
        elif isinstance(in_.rules, DMSInputRules):
            return DMSRules
        else:
            raise NeatTypeError(f"Unsupported rules type: {type(in_)}")

    def _get_validation_cls(self, rules: VerifiedRules) -> type:
        if isinstance(rules, InformationRules):
            return InformationValidation
        elif isinstance(rules, DMSRules):
            return DMSValidation
        else:
            raise NeatTypeError(f"Unsupported rules type: {type(rules)}")



import sys
import warnings
import zipfile
from collections import ChainMap
from collections.abc import Iterable, MutableMapping
from dataclasses import Field, dataclass, field, fields
from pathlib import Path
from typing import Any, ClassVar, Literal, cast

import yaml
from cognite.client import data_modeling as dm
from cognite.client.data_classes import DatabaseWrite, TransformationWrite
from cognite.client.data_classes.data_modeling.views import (
    ReverseDirectRelationApply,
    SingleEdgeConnection,
    SingleEdgeConnectionApply,
    SingleReverseDirectRelation,
    SingleReverseDirectRelationApply,
    ViewProperty,
    ViewPropertyApply,
)

from cognite.neat._client.data_classes.data_modeling import (
    CogniteResourceDict,
    ContainerApplyDict,
    NodeApplyDict,
    SpaceApplyDict,
    ViewApplyDict,
)
from cognite.neat._issues.errors import (
    NeatYamlError,
)
from cognite.neat._issues.warnings import (
    FileTypeUnexpectedWarning,
    ResourcesDuplicatedWarning,
)
from cognite.neat._utils.text import to_camel_case

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class DMSSchema:
    data_model: dm.DataModelApply | None = None
    spaces: SpaceApplyDict = field(default_factory=SpaceApplyDict)
    views: ViewApplyDict = field(default_factory=ViewApplyDict)
    containers: ContainerApplyDict = field(default_factory=ContainerApplyDict)
    node_types: NodeApplyDict = field(default_factory=NodeApplyDict)

    _FIELD_NAME_BY_RESOURCE_TYPE: ClassVar[dict[str, str]] = {
        "container": "containers",
        "view": "views",
        "datamodel": "data_model",
        "space": "spaces",
        "node": "node_types",
    }

    @classmethod
    def from_directory(cls, directory: str | Path) -> Self:
        """Load a schema from a directory containing YAML files.

        The directory is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data, context = cls._read_directory(Path(directory))
        return cls.load(data, context)

    @classmethod
    def _read_directory(cls, directory: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data: dict[str, Any] = {}
        context: dict[str, list[Path]] = {}
        for yaml_file in directory.rglob("*.yaml"):
            if "." in yaml_file.stem:
                resource_type = yaml_file.stem.rsplit(".", 1)[-1]
                if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type):
                    data.setdefault(attr_name, [])
                    context.setdefault(attr_name, [])
                    try:
                        loaded = yaml.safe_load(yaml_file.read_text())
                    except Exception as e:
                        warnings.warn(
                            FileTypeUnexpectedWarning(yaml_file, frozenset([".yaml", ".yml"]), str(e)), stacklevel=2
                        )
                        continue

                    if isinstance(loaded, list):
                        data[attr_name].extend(loaded)
                        context[attr_name].extend([yaml_file] * len(loaded))
                    else:
                        data[attr_name].append(loaded)
                        context[attr_name].append(yaml_file)
        return data, context

    def to_directory(
        self,
        directory: str | Path,
        exclude: set[str] | None = None,
        new_line: str | None = "\n",
        encoding: str | None = "utf-8",
    ) -> None:
        """Save the schema to a directory as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            directory (str | Path): The directory to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
            new_line (str): The line endings to use in the output files. Defaults to "\n".
            encoding (str): The encoding to use in the output files. Defaults to "utf-8".
        """
        path_dir = Path(directory)
        exclude_set = exclude or set()
        data_models = path_dir / "data_models"
        data_models.mkdir(parents=True, exist_ok=True)
        if "spaces" not in exclude_set:
            for space in self.spaces.values():
                (data_models / f"{space.space}.space.yaml").write_text(
                    space.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "data_models" not in exclude_set and self.data_model:
            (data_models / f"{self.data_model.external_id}.datamodel.yaml").write_text(
                self.data_model.dump_yaml(), newline=new_line, encoding=encoding
            )
        if "views" not in exclude_set and self.views:
            view_dir = data_models / "views"
            view_dir.mkdir(parents=True, exist_ok=True)
            for view in self.views.values():
                (view_dir / f"{view.external_id}.view.yaml").write_text(
                    view.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "containers" not in exclude_set and self.containers:
            container_dir = data_models / "containers"
            container_dir.mkdir(parents=True, exist_ok=True)
            for container in self.containers.values():
                (container_dir / f"{container.external_id}.container.yaml").write_text(
                    container.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "node_types" not in exclude_set and self.node_types:
            node_dir = data_models / "nodes"
            node_dir.mkdir(parents=True, exist_ok=True)
            for node in self.node_types.values():
                (node_dir / f"{node.external_id}.node.yaml").write_text(
                    node.dump_yaml(), newline=new_line, encoding=encoding
                )

    @classmethod
    def from_zip(cls, zip_file: str | Path) -> Self:
        """Load a schema from a ZIP file containing YAML files.

        The ZIP file is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data, context = cls._read_zip(Path(zip_file))
        return cls.load(data, context)

    @classmethod
    def _read_zip(cls, zip_file: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data: dict[str, list[Any]] = {}
        context: dict[str, list[Path]] = {}
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename.endswith(".yaml"):
                    if "/" not in file_info.filename:
                        continue
                    filename = Path(file_info.filename.split("/")[-1])
                    if "." not in filename.stem:
                        continue
                    resource_type = filename.stem.rsplit(".", 1)[-1]
                    if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type.casefold()):
                        data.setdefault(attr_name, [])
                        context.setdefault(attr_name, [])
                        try:
                            loaded = yaml.safe_load(zip_ref.read(file_info).decode())
                        except Exception as e:
                            warnings.warn(
                                FileTypeUnexpectedWarning(filename, frozenset([".yaml", ".yml"]), str(e)), stacklevel=2
                            )
                            continue
                        if isinstance(loaded, list):
                            data[attr_name].extend(loaded)
                            context[attr_name].extend([filename] * len(loaded))
                        else:
                            data[attr_name].append(loaded)
                            context[attr_name].append(filename)
        return data, context

    def to_zip(self, zip_file: str | Path, exclude: set[str] | None = None) -> None:
        """Save the schema to a ZIP file as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            zip_file (str | Path): The ZIP file to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
        """
        exclude_set = exclude or set()
        with zipfile.ZipFile(zip_file, "w") as zip_ref:
            if "spaces" not in exclude_set:
                for space in self.spaces.values():
                    zip_ref.writestr(f"data_models/{space.space}.space.yaml", space.dump_yaml())
            if "data_models" not in exclude_set and self.data_model:
                zip_ref.writestr(
                    f"data_models/{self.data_model.external_id}.datamodel.yaml", self.data_model.dump_yaml()
                )
            if "views" not in exclude_set:
                for view in self.views.values():
                    zip_ref.writestr(f"data_models/views/{view.external_id}.view.yaml", view.dump_yaml())
            if "containers" not in exclude_set:
                for container in self.containers.values():
                    zip_ref.writestr(
                        f"data_models/containers/{container.external_id}.container.yaml", container.dump_yaml()
                    )
            if "node_types" not in exclude_set:
                for node in self.node_types.values():
                    zip_ref.writestr(f"data_models/nodes/{node.external_id}.node.yaml", node.dump_yaml())

    @classmethod
    def load(cls, data: str | dict[str, list[Any]], context: dict[str, list[Path]] | None = None) -> Self:
        """Loads a schema from a dictionary or a YAML or JSON formatted string.

        Args:
            data: The data to load the schema from. This can be a dictionary, a YAML or JSON formatted string.
            context: This provides linage for where the data was loaded from. This is used in Warnings
                if a single item fails to load.

        Returns:
            DMSSchema: The loaded schema.
        """
        context = context or {}
        if isinstance(data, str):
            # YAML is a superset of JSON, so we can use the same parser
            try:
                data_dict = yaml.safe_load(data)
            except Exception as e:
                raise NeatYamlError(str(e)) from None
            if not isinstance(data_dict, dict) and all(isinstance(v, list) for v in data_dict.values()):
                raise NeatYamlError(f"Invalid data structure: {type(data)}", "dict[str, list[Any]]") from None
        else:
            data_dict = data
        loaded: dict[str, Any] = {}
        for attr in fields(cls):
            if items := data_dict.get(attr.name) or data_dict.get(to_camel_case(attr.name)):
                if attr.name == "data_model":
                    if isinstance(items, list) and len(items) > 1:
                        try:
                            data_model_ids = [dm.DataModelId.load(item) for item in items]
                        except Exception as e:
                            data_model_file = context.get(attr.name, [Path("UNKNOWN")])[0]
                            warnings.warn(
                                FileTypeUnexpectedWarning(
                                    data_model_file, frozenset([dm.DataModelApply.__name__]), str(e)
                                ),
                                stacklevel=2,
                            )
                        else:
                            warnings.warn(
                                ResourcesDuplicatedWarning(
                                    frozenset(data_model_ids),
                                    "data model",
                                    "Will use the first DataModel.",
                                ),
                                stacklevel=2,
                            )
                    item = items[0] if isinstance(items, list) else items
                    try:
                        loaded[attr.name] = dm.DataModelApply.load(item)
                    except Exception as e:
                        data_model_file = context.get(attr.name, [Path("UNKNOWN")])[0]
                        warnings.warn(
                            FileTypeUnexpectedWarning(data_model_file, frozenset([dm.DataModelApply.__name__]), str(e)),
                            stacklevel=2,
                        )
                else:
                    try:
                        loaded[attr.name] = attr.type.load(items)  # type: ignore[union-attr]
                    except Exception as e:
                        loaded[attr.name] = cls._load_individual_resources(
                            items, attr, str(e), context.get(attr.name, [])
                        )
        return cls(**loaded)

    @classmethod
    def _load_individual_resources(cls, items: list, attr: Field, trigger_error: str, resource_context) -> list[Any]:
        type_ = cast(type, attr.type)
        resources = type_([])
        if not hasattr(type_, "_RESOURCE"):
            warnings.warn(
                FileTypeUnexpectedWarning(Path("UNKNOWN"), frozenset([type_.__name__]), trigger_error), stacklevel=2
            )
            return resources
        # Fallback to load individual resources.
        single_cls = type_._RESOURCE
        for no, item in enumerate(items):
            try:
                loaded_instance = single_cls.load(item)
            except Exception as e:
                try:
                    filepath = resource_context[no]
                except IndexError:
                    filepath = Path("UNKNOWN")
                # We use repr(e) instead of str(e) to include the exception type in the warning message
                warnings.warn(
                    FileTypeUnexpectedWarning(filepath, frozenset([single_cls.__name__]), repr(e)), stacklevel=2
                )
            else:
                resources.append(loaded_instance)
        return resources

    def dump(self, camel_case: bool = True, sort: bool = True) -> dict[str, Any]:
        """Dump the schema to a dictionary that can be serialized to JSON.

        Args:
            camel_case (bool): If True, the keys in the output dictionary will be in camel case.
            sort (bool): If True, the items in the output dictionary will be sorted by their ID.
                This is useful for deterministic output which is useful for comparing schemas.

        Returns:
            dict: The schema as a dictionary.
        """
        output: dict[str, Any] = {}
        cls_fields = sorted(fields(self), key=lambda f: f.name) if sort else fields(self)
        for attr in cls_fields:
            if items := getattr(self, attr.name):
                key = to_camel_case(attr.name) if camel_case else attr.name
                if isinstance(items, CogniteResourceDict):
                    if sort:
                        output[key] = [
                            item.dump(camel_case) for item in sorted(items.values(), key=self._to_sortable_identifier)
                        ]
                    else:
                        output[key] = items.dump(camel_case)
                else:
                    output[key] = items.dump(camel_case=camel_case)
        return output

    @classmethod
    def _to_sortable_identifier(cls, item: Any) -> str | tuple[str, str] | tuple[str, str, str]:
        if isinstance(item, dm.ContainerApply | dm.ViewApply | dm.DataModelApply | dm.NodeApply):
            identifier = item.as_id().as_tuple()
            if len(identifier) == 3 and identifier[2] is None:
                return identifier[:2]  # type: ignore[misc]
            return cast(tuple[str, str] | tuple[str, str, str], identifier)
        elif isinstance(item, dm.SpaceApply):
            return item.space
        elif isinstance(item, TransformationWrite):
            return item.external_id or ""
        elif isinstance(item, DatabaseWrite):
            return item.name or ""
        else:
            raise ValueError(f"Cannot sort item of type {type(item)}")

    def referenced_spaces(self, include_indirect_references: bool = True) -> set[str]:
        """Get the spaces referenced by the schema.

        Args:
            include_indirect_references (bool): If True, the spaces referenced by as view.implements, and
                view.referenced_containers will be included in the output.
        Returns:
            set[str]: The spaces referenced by the schema.
        """
        referenced_spaces = {view.space for view in self.views.values()}
        referenced_spaces |= {container.space for container in self.containers.values()}
        if include_indirect_references:
            referenced_spaces |= {
                container.space for view in self.views.values() for container in view.referenced_containers()
            }
            referenced_spaces |= {parent.space for view in self.views.values() for parent in view.implements or []}
        referenced_spaces |= {node.space for node in self.node_types.values()}
        if self.data_model:
            referenced_spaces |= {self.data_model.space}
            referenced_spaces |= {view.space for view in self.data_model.views or []}
        referenced_spaces |= {s.space for s in self.spaces.values()}
        return referenced_spaces

    def referenced_container(self) -> set[dm.ContainerId]:
        referenced_containers = {
            container for view in self.views.values() for container in view.referenced_containers()
        }
        referenced_containers |= set(self.containers.keys())
        return referenced_containers

    def externally_referenced_containers(self) -> set[dm.ContainerId]:
        """Get the containers referenced by the schema that are not defined in the schema."""
        return {container for container in self.referenced_container() if container not in self.containers}

    def as_read_model(self) -> dm.DataModel[dm.View]:
        if self.data_model is None:
            raise ValueError("Data model is not defined")
        all_containers = self.containers.copy()
        all_views = self.views.copy()
        views: list[dm.View] = []
        for view in self.views.values():
            referenced_containers = ContainerApplyDict()
            properties: dict[str, ViewProperty] = {}
            # ChainMap is used to merge properties from the view and its parents
            # Note that the order of the ChainMap is important, as the first dictionary has the highest priority
            # So if a child and parent have the same property, the child property will be used.
            write_properties = ChainMap(view.properties, *(all_views[v].properties for v in view.implements or []))  # type: ignore[arg-type]
            for prop_name, prop in write_properties.items():
                read_prop = self._as_read_properties(prop, all_containers)
                if isinstance(read_prop, dm.MappedProperty) and read_prop.container not in referenced_containers:
                    referenced_containers[read_prop.container] = all_containers[read_prop.container]
                properties[prop_name] = read_prop

            read_view = dm.View(
                space=view.space,
                external_id=view.external_id,
                version=view.version,
                description=view.description,
                name=view.name,
                filter=view.filter,
                implements=view.implements.copy(),
                used_for=self._used_for(referenced_containers.values()),
                writable=self._writable(properties.values(), referenced_containers.values()),
                properties=properties,
                is_global=False,
                last_updated_time=0,
                created_time=0,
            )
            views.append(read_view)

        return dm.DataModel(
            space=self.data_model.space,
            external_id=self.data_model.external_id,
            version=self.data_model.version,
            name=self.data_model.name,
            description=self.data_model.description,
            views=views,
            is_global=False,
            last_updated_time=0,
            created_time=0,
        )

    @staticmethod
    def _as_read_properties(
        write: ViewPropertyApply, all_containers: MutableMapping[dm.ContainerId, dm.ContainerApply]
    ) -> ViewProperty:
        if isinstance(write, dm.MappedPropertyApply):
            container_prop = all_containers[write.container].properties[write.container_property_identifier]
            return dm.MappedProperty(
                container=write.container,
                container_property_identifier=write.container_property_identifier,
                name=write.name,
                description=write.description,
                source=write.source,
                type=container_prop.type,
                nullable=container_prop.nullable,
                auto_increment=container_prop.auto_increment,
                immutable=container_prop.immutable,
                # Likely bug in SDK.
                default_value=container_prop.default_value,  # type: ignore[arg-type]
            )
        if isinstance(write, dm.EdgeConnectionApply):
            edge_cls = SingleEdgeConnection if isinstance(write, SingleEdgeConnectionApply) else dm.MultiEdgeConnection
            return edge_cls(
                type=write.type,
                source=write.source,
                name=write.name,
                description=write.description,
                edge_source=write.edge_source,
                direction=write.direction,
            )
        if isinstance(write, ReverseDirectRelationApply):
            relation_cls = (
                SingleReverseDirectRelation
                if isinstance(write, SingleReverseDirectRelationApply)
                else dm.MultiReverseDirectRelation
            )
            return relation_cls(
                source=write.source,
                through=write.through,
                name=write.name,
                description=write.description,
            )
        raise ValueError(f"Cannot convert {write} to read format")

    @staticmethod
    def _used_for(containers: Iterable[dm.ContainerApply]) -> Literal["node", "edge", "all"]:
        used_for = {container.used_for for container in containers}
        if used_for == {"node"}:
            return "node"
        if used_for == {"edge"}:
            return "edge"
        return "all"

    @staticmethod
    def _writable(properties: Iterable[ViewProperty], containers: Iterable[dm.ContainerApply]) -> bool:
        used_properties = {
            (prop.container, prop.container_property_identifier)
            for prop in properties
            if isinstance(prop, dm.MappedProperty)
        }
        required_properties = {
            (container.as_id(), prop_id)
            for container in containers
            for prop_id, prop in container.properties.items()
            if not prop.nullable
        }
        # If a container has a required property that is not used by the view, the view is not writable
        return not bool(required_properties - used_properties)



import warnings
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from pathlib import Path

from ._base import IssueList, NeatError, NeatWarning

__all__ = ["FORMATTER_BY_NAME", "BasicHTML", "Formatter"]


class Formatter(ABC):
    file_suffix: str
    default_file_prefix: str = "validation_report"

    @abstractmethod
    def create_report(self, issues: IssueList) -> str:
        raise NotImplementedError()

    @property
    def default_file_name(self) -> str:
        return f"{self.default_file_prefix}_{type(self).__name__.lower()}{self.file_suffix}"

    def write_to_file(self, issues: IssueList, file_or_dir_path: Path | None = None) -> None:
        if file_or_dir_path is None:
            file_or_dir_path = Path(self.default_file_name)
        elif file_or_dir_path.is_dir():
            file_or_dir_path = file_or_dir_path / self.default_file_name

        if file_or_dir_path.suffix != self.file_suffix:
            warnings.warn(
                f"File suffix is not {self.file_suffix}. Appending suffix to file path.", UserWarning, stacklevel=2
            )
            file_or_dir_path = file_or_dir_path.with_suffix(self.file_suffix)
        file_or_dir_path.write_text(self.create_report(issues))


class BasicHTML(Formatter):
    file_suffix = ".html"

    def __init__(self):
        self._doc = ET.Element("html")
        self._body = ET.SubElement(self._doc, "body")

    def create_report(self, issues: IssueList) -> str:
        errors = [issue for issue in issues if isinstance(issue, NeatError)]
        warnings_ = [issue for issue in issues if isinstance(issue, NeatWarning)]
        self._doc.clear()
        self._body = ET.SubElement(self._doc, "body")
        h1 = ET.SubElement(self._body, "h1")
        h1.text = f"Validation Report: {issues.title or 'Missing title'}"

        if errors:
            h2 = ET.SubElement(self._body, "h2")
            h2.text = "Errors"
            self._write_errors_or_warnings(errors)

        if warnings_:
            h2 = ET.SubElement(self._body, "h2")
            h2.text = "Warnings"
            self._write_errors_or_warnings(warnings_)

        return ET.tostring(self._doc, encoding="unicode")

    def _write_errors_or_warnings(self, issues: list[NeatError] | list[NeatWarning]) -> None:
        issue_name = "errors" if isinstance(issues[0], NeatError) else "warnings"
        main_categories = {base_ for issue in issues for base_ in type(issue).__bases__}
        for category in main_categories:
            issues_in_category: list[NeatError] | list[NeatWarning] = [  # type: ignore[assignment]
                issue for issue in issues if isinstance(issue, category)
            ]
            h3 = ET.SubElement(self._body, "h3")
            h3.text = category.__name__
            p = ET.SubElement(self._body, "p")
            p.text = f"Total: {len(issues_in_category)} {issue_name}"
            ul = ET.SubElement(self._body, "ul")
            for issue in issues_in_category:
                li = ET.SubElement(ul, "li")
                li.text = issue.as_message()


FORMATTER_BY_NAME: dict[str, type[Formatter]] = {
    subclass.__name__: subclass  # type: ignore[type-abstract]
    for subclass in Formatter.__subclasses__()
    if ABC not in subclass.__bases__  # type: ignore[type-abstract]
}



from dataclasses import dataclass
from typing import Generic

from cognite.neat._constants import DMS_DIRECT_RELATION_LIST_LIMIT
from cognite.neat._issues._base import ResourceType

from ._resources import ResourceNeatWarning, T_Identifier, T_ReferenceIdentifier


@dataclass(unsafe_hash=True)
class PropertyWarning(ResourceNeatWarning[T_Identifier]):
    """Base class for property warnings {resource_type} with identifier {identifier}.{property_name}"""

    property_name: str


@dataclass(unsafe_hash=True)
class PropertyTypeNotSupportedWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name}
    of unsupported type {property_type}. This will be ignored."""

    property_type: str


@dataclass(unsafe_hash=True)
class PropertyNotFoundWarning(PropertyWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} does not have a property {property_name} referred
    to by {referred_type} {referred_by} does not exist. This will be ignored.
    """

    fix = "Ensure the {resource_type} {identifier} has a {property_name} property"
    referred_by: T_ReferenceIdentifier
    referred_type: ResourceType


@dataclass(unsafe_hash=True)
class PropertyDefinitionDuplicatedWarning(PropertyWarning[T_Identifier]):
    """Got multiple values for the {resource_type} {identifier}.{property_name} {values}.
    {default_action}"""

    extra = "Recommended action: {recommended_action}"

    values: frozenset[str]
    default_action: str
    recommended_action: str | None = None


@dataclass(unsafe_hash=True)
class PropertyValueTypeUndefinedWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name}
    which has undefined value type. This may result in unexpected behavior when exporting rules.
    {default_action}"""

    extra = "Recommended action: {recommended_action}"

    default_action: str
    recommended_action: str | None = None


@dataclass(unsafe_hash=True)
class PropertyOverwritingWarning(PropertyWarning[T_Identifier]):
    """Overwriting the {overwriting} for {property_name} in the {resource_type}
    with identifier {identifier}."""

    overwriting: tuple[str, ...]


@dataclass(unsafe_hash=True)
class PropertyDataTypeConversionWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} failed to convert the property {property_name}: {error}"""

    error: str


@dataclass(unsafe_hash=True)
class PropertyDirectRelationLimitWarning(PropertyWarning[T_Identifier]):
    """The listable direct relation property {property_name} in the {resource_type} with identifier {identifier}
    has more than {limit} relations. The relations will be sorted by (space, externalId) and truncated."""

    resource_type = "view"

    limit: int = DMS_DIRECT_RELATION_LIST_LIMIT


@dataclass(unsafe_hash=True)
class PropertyMultipleValueWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name} with multiple values.
    Selecting the first value {value}, the rest will be ignored."""

    value: str



import os
import re
import shutil
import sys
import tempfile
import warnings
from collections.abc import Callable
from pathlib import Path
from typing import Literal, cast

from cognite.client import CogniteClient
from packaging.version import Version
from packaging.version import parse as parse_version

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._version import __engine__

ENVIRONMENT_VARIABLE = "NEATENGINE"
PACKAGE_NAME = "neatengine"
PYVERSION = f"{sys.version_info.major}{sys.version_info.minor}"


def load_neat_engine(client: CogniteClient | None, location: Literal["newest", "cache"]) -> str | None:
    if location not in ["newest", "cache"]:
        raise NeatValueError(f"Cannot load engine from location: {location}")

    if not __engine__.startswith("^"):
        # Using value error as this is a developer error
        raise ValueError(f"Invalid engine version: {__engine__}")

    lower_bound = parse_version(__engine__[1:])
    upper_bound = Version(f"{lower_bound.major + 1}.0.0")

    cache_dir = Path(tempfile.gettempdir()) / PACKAGE_NAME
    cache_dir.mkdir(exist_ok=True)
    pattern = re.compile(rf"{PACKAGE_NAME}-(\d+\.\d+\.\d+)-{PYVERSION}.zip")

    candidates: dict[Version, Callable[[], Path]] = {}
    if location == "cache" and cache_dir.exists():
        candidates = _load_from_path(cache_dir, pattern, lower_bound, upper_bound)

    if location == "newest" or not candidates:
        # Loading in reverse order of priority
        # 4. Downloads folder
        candidates = _load_from_path(Path.home() / "Downloads", pattern, lower_bound, upper_bound)
        # 3. Current working directory
        candidates.update(_load_from_path(Path.cwd(), pattern, lower_bound, upper_bound))
        # 2. CDF
        if client:
            candidates.update(_load_from_cdf(client, pattern, lower_bound, upper_bound, cache_dir))
        # 1. Environment variable
        if ENVIRONMENT_VARIABLE in os.environ:
            environ_path = Path(os.environ[ENVIRONMENT_VARIABLE])
            if environ_path.exists():
                candidates.update(_load_from_path(environ_path, pattern, lower_bound, upper_bound))
            else:
                warnings.warn(
                    f"Environment variable {ENVIRONMENT_VARIABLE} points to non-existing path: {environ_path}",
                    UserWarning,
                    stacklevel=2,
                )

    if not candidates:
        return None

    selected_version = max(candidates.keys(), default=None)
    if not selected_version:
        return None
    source_path = candidates[selected_version]()
    destination_path = cache_dir / source_path.name
    if not destination_path.exists():
        shutil.copy(source_path, destination_path)
    sys.path.append(str(destination_path))
    try:
        from neatengine._version import __version__ as engine_version  # type: ignore[import-not-found]
    except ImportError:
        return None
    return engine_version


def _load_from_path(
    path: Path, pattern: re.Pattern[str], lower_bound: Version, upper_bound: Version
) -> dict[Version, Callable[[], Path]]:
    if path.is_file() and (match := pattern.match(path.name)):
        version = parse_version(match.group(1))
        if lower_bound <= version < upper_bound:
            return {parse_version(match.group(1)): lambda: path}
        return {}
    elif path.is_dir():
        output: dict[Version, Callable[[], Path]] = {}
        for candidate in path.iterdir():
            if candidate.is_file() and (match := pattern.match(candidate.name)):
                version = parse_version(match.group(1))
                if lower_bound <= version < upper_bound:
                    # Setting default value to ensure we use the candidate from the current iteration
                    # If not set, the function will use the last candidate from the loop
                    def return_path(the_path: Path = candidate) -> Path:
                        return the_path

                    output[parse_version(match.group(1))] = return_path

        return output
    return {}


def _load_from_cdf(
    client: CogniteClient, pattern: re.Pattern[str], lower_bound: Version, upper_bound: Version, cache_dir: Path
) -> dict[Version, Callable[[], Path]]:
    file_metadata = client.files.list(
        limit=-1,
        data_set_external_ids=PACKAGE_NAME,
        external_id_prefix=PACKAGE_NAME,
        metadata={"python_version": PYVERSION},
    )
    output: dict[Version, Callable[[], Path]] = {}
    for file in file_metadata:
        name = cast(str, file.name)

        # Use function to lazily download file
        # Setting default value to ensure we use the file_id from the current iteration
        # If not set, the function will use the last file_id from the loop
        def download_file(file_id: int = file.id, filename: str = name) -> Path:
            client.files.download(cache_dir, file_id)
            return cache_dir / filename

        if match := pattern.match(name):
            version = parse_version(match.group(1))
            if lower_bound <= version < upper_bound:
                output[version] = download_file

    return output



from collections.abc import Iterable
from typing import Any, Protocol

from rdflib import Literal, URIRef


class Extractor(Protocol):
    def extract(self) -> Iterable[tuple[URIRef, URIRef, Literal | URIRef]]: ...


class ConfigAPI(Protocol):
    format: str | None = None
    file: Any | None
    type: str | None
    primary_key: str | None
    mapping: Any | None = None


class NeatEngine(Protocol):
    version: str = "2.0.0"

    @property
    def set(self) -> ConfigAPI: ...

    def create_extractor(self) -> Extractor: ...



import re

# Spaces are allowed, but we replace them as well
_ILLEGAL_CHARACTERS = re.compile(r"[<>:\"/\\|?*\s]")


def to_directory_compatible(text: str) -> str:
    """Convert a string to be compatible with directory names on all platforms"""
    cleaned = _ILLEGAL_CHARACTERS.sub("_", text)
    # Replace multiple underscores with a single one
    return re.sub(r"_+", "_", cleaned)



import warnings
from collections.abc import Iterator
from contextlib import contextmanager

from pydantic import ValidationError

from cognite.neat._utils.spreadsheet import SpreadsheetRead

from ._base import IssueList, MultiValueError, NeatError
from ._factory import from_pydantic_errors, from_warning


@contextmanager
def catch_warnings() -> Iterator[IssueList]:
    """Catch warnings and append them to the issues list."""
    issues = IssueList()
    with warnings.catch_warnings(record=True) as warning_logger:
        warnings.simplefilter("always")
        try:
            yield issues
        finally:
            if warning_logger:
                issues.extend([from_warning(warning) for warning in warning_logger])


@contextmanager
def catch_issues(read_info_by_sheet: dict[str, SpreadsheetRead] | None = None) -> Iterator[IssueList]:
    """This is an internal help function to handle issues and warnings.

    Args:
        read_info_by_sheet (dict[str, SpreadsheetRead]): The read information by sheet. This is used to adjust
            the row numbers in the errors/warnings.

    Returns:
        IssueList: The list of issues.

    """
    with catch_warnings() as issues:
        try:
            yield issues
        except ValidationError as e:
            issues.extend(from_pydantic_errors(e.errors(), read_info_by_sheet))
        except NeatError as single:
            issues.append(single)
        except MultiValueError as multi:
            issues.extend(multi.errors)



import os
import platform
import tempfile
import threading
import uuid
from contextlib import suppress
from functools import cached_property
from pathlib import Path
from typing import Any

from mixpanel import Consumer, Mixpanel  # type: ignore[import-untyped]

from cognite.neat._constants import IN_NOTEBOOK, IN_PYODIDE
from cognite.neat._version import __version__

_NEAT_MIXPANEL_TOKEN: str = "bd630ad149d19999df3989c3a3750c4f"


class Collector:
    def __init__(self, skip_tracking: bool = False) -> None:
        self.mp = Mixpanel(_NEAT_MIXPANEL_TOKEN, consumer=Consumer(api_host="api-eu.mixpanel.com"))
        tmp_dir = Path(tempfile.gettempdir()).resolve()
        self._opt_status_file = tmp_dir / "neat-opt-status.bin"
        self._distinct_id_file = tmp_dir / "neat-distinct-id.bin"
        self.skip_tracking = self.opted_out or skip_tracking

    @cached_property
    def _opt_status(self) -> str:
        if self._opt_status_file.exists():
            return self._opt_status_file.read_text()
        return ""

    def _bust_opt_status(self) -> None:
        self.__dict__.pop("_opt_status", None)

    @property
    def opted_out(self) -> bool:
        return self._opt_status == "opted-out"

    @property
    def opted_in(self) -> bool:
        return self._opt_status == "opted-in"

    @staticmethod
    def _get_environment() -> str:
        if IN_PYODIDE:
            return "pyodide"
        if IN_NOTEBOOK:
            return "notebook"
        return "python"

    def track_session_command(self, command: str, *args, **kwargs) -> None:
        event_information = {
            "neatVersion": __version__,
            "$os": platform.system(),
            "pythonVersion": platform.python_version(),
            "environment": self._get_environment(),
        }

        if len(args) > 1:
            # The first argument is self.
            for i, arg in enumerate(args[1:]):
                event_information[f"arg{i}"] = self._serialize_value(arg)[:500]

        if kwargs:
            for key, value in kwargs.items():
                event_information[key] = self._serialize_value(value)[:500]

        with suppress(RuntimeError):
            # In case any thread issues, the tracking should not crash the program
            self._track(command, event_information)

    @staticmethod
    def _serialize_value(value: Any) -> str:
        if isinstance(value, (str | int | float | bool)):
            return str(value)
        if isinstance(value, list | tuple | dict):
            return str(value)
        if callable(value):
            return value.__name__
        return str(type(value))

    def _track(self, event_name: str, event_information: dict[str, Any]) -> bool:
        if self.skip_tracking or not self.opted_in or "PYTEST_CURRENT_TEST" in os.environ:
            return False

        distinct_id = self.get_distinct_id()

        def track() -> None:
            # If we are unable to connect to Mixpanel, we don't want to crash the program
            with suppress(ConnectionError):
                self.mp.track(
                    distinct_id,
                    event_name,
                    event_information,
                )

        thread = threading.Thread(
            target=track,
            daemon=False,
        )
        thread.start()
        return True

    def get_distinct_id(self) -> str:
        if self._distinct_id_file.exists():
            return self._distinct_id_file.read_text()

        distinct_id = f"{platform.system()}-{platform.python_version()}-{uuid.uuid4()!s}"
        self._distinct_id_file.write_text(distinct_id)
        with suppress(ConnectionError):
            self.mp.people_set(
                distinct_id,
                {
                    "$os": platform.system(),
                    "$python_version": platform.python_version(),
                    "$distinct_id": distinct_id,
                },
            )
        return distinct_id

    def enable(self) -> None:
        self._opt_status_file.write_text("opted-in")
        self._bust_opt_status()

    def disable(self) -> None:
        self._opt_status_file.write_text("opted-out")
        self._bust_opt_status()


_COLLECTOR = Collector()



from functools import lru_cache
from pathlib import Path

import yaml

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.models.dms import DMSInputRules, DMSRules

_CLASSIC_TO_CORE_MAPPING = Path(__file__).resolve().parent / "_classic2core.yaml"


@lru_cache(maxsize=1)
def _read_source_file() -> str:
    return _CLASSIC_TO_CORE_MAPPING.read_text()


def load_classic_to_core_mapping(org_name: str | None, source_space: str, source_version: str) -> DMSRules:
    from cognite.neat._rules.importers import YAMLImporter
    from cognite.neat._rules.transformers import VerifyDMSRules

    raw_str = _read_source_file()
    if org_name is not None:
        raw_str = raw_str.replace("Classic", org_name)

    loaded = yaml.safe_load(raw_str)
    loaded["metadata"]["space"] = source_space
    loaded["metadata"]["version"] = source_version

    read: ReadRules[DMSInputRules] = YAMLImporter(loaded).to_rules()
    if not isinstance(read.rules, DMSInputRules):
        raise NeatValueError(f"Expected DMS rules, but got {type(read.rules).__name__}")

    verified = VerifyDMSRules(validate=False).transform(read)

    return verified



from typing import Annotated, Any

from pydantic import (
    AnyHttpUrl,
    BeforeValidator,
    PlainSerializer,
)

from ._single_value import AssetEntity, ClassEntity, ContainerEntity, RelationshipEntity, ViewEntity


def _split_str(v: Any) -> list[str]:
    if isinstance(v, str):
        return v.replace(", ", ",").split(",")
    return v


def _join_str(v: list[ClassEntity]) -> str | None:
    return ",".join([entry.id for entry in v]) if v else None


def _generate_cdf_resource_list(v: Any) -> list[AssetEntity | RelationshipEntity]:
    results = []
    for item in _split_str(v):
        if isinstance(item, str):
            if "relationship" in item.lower():
                results.append(RelationshipEntity.load(item))
            elif "asset" in item.lower():
                results.append(AssetEntity.load(item))  # type: ignore
            else:
                raise ValueError(f"Unsupported implementation definition: {item}")

        elif isinstance(item, AssetEntity | RelationshipEntity):
            results.append(item)
        else:
            raise ValueError(f"Unsupported implementation definition: {item}")

    return results  # type: ignore


ClassEntityList = Annotated[
    list[ClassEntity],
    BeforeValidator(_split_str),
]


CdfResourceEntityList = Annotated[
    list[AssetEntity | RelationshipEntity],
    BeforeValidator(_generate_cdf_resource_list),
    PlainSerializer(
        _join_str,
        return_type=str,
        when_used="unless-none",
    ),
]


ContainerEntityList = Annotated[
    list[ContainerEntity],
    BeforeValidator(_split_str),
]

ViewEntityList = Annotated[
    list[ViewEntity],
    BeforeValidator(_split_str),
    PlainSerializer(
        _join_str,
        return_type=str,
        when_used="unless-none",
    ),
]

URLEntity = Annotated[
    AnyHttpUrl,
    PlainSerializer(lambda v: str(v), return_type=str, when_used="unless-none"),
]



import itertools
from collections import Counter, defaultdict

from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.errors._resources import (
    ResourceDuplicatedError,
    ResourceNotDefinedError,
)
from cognite.neat._issues.warnings._models import UndefinedClassWarning
from cognite.neat._issues.warnings._resources import (
    ResourceNotDefinedWarning,
    ResourceRegexViolationWarning,
)
from cognite.neat._rules._constants import PATTERNS, EntityTypes
from cognite.neat._rules.models.entities import ClassEntity, UnknownEntity
from cognite.neat._rules.models.entities._multi_value import MultiValueTypeInfo
from cognite.neat._utils.spreadsheet import SpreadsheetRead
from cognite.neat._utils.text import humanize_collection

from ._rules import InformationRules


class InformationValidation:
    """This class does all the validation of the Information rules that have dependencies
    between components."""

    def __init__(self, rules: InformationRules, read_info_by_spreadsheet: dict[str, SpreadsheetRead] | None = None):
        self.rules = rules
        self._read_info_by_spreadsheet = read_info_by_spreadsheet or {}
        self._metadata = rules.metadata
        self._properties = rules.properties
        self._classes = rules.classes
        self.issue_list = IssueList()

    def validate(self) -> IssueList:
        self._duplicated_resources()
        self._namespaces_reassigned()
        self._classes_without_properties()
        self._undefined_classes()
        self._parent_class_defined()
        self._referenced_classes_exist()
        self._referenced_value_types_exist()
        self._regex_compliance_with_dms()

        return self.issue_list

    def _duplicated_resources(self) -> None:
        properties_sheet = self._read_info_by_spreadsheet.get("Properties")
        classes_sheet = self._read_info_by_spreadsheet.get("Classes")

        visited = defaultdict(list)
        for row_no, property_ in enumerate(self._properties):
            visited[property_._identifier()].append(
                properties_sheet.adjusted_row_number(row_no) if properties_sheet else row_no + 1
            )

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            self.issue_list.append(
                ResourceDuplicatedError(
                    identifier[1],
                    "property",
                    (
                        "the Properties sheet at row "
                        f"{humanize_collection(rows)}"
                        " if data model is read from a spreadsheet."
                    ),
                )
            )

        visited = defaultdict(list)
        for row_no, class_ in enumerate(self._classes):
            visited[class_._identifier()].append(
                classes_sheet.adjusted_row_number(row_no) if classes_sheet else row_no + 1
            )

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            self.issue_list.append(
                ResourceDuplicatedError(
                    identifier[0],
                    "class",
                    (f"the Classes sheet at row {humanize_collection(rows)} if data model is read from a spreadsheet."),
                )
            )

    def _classes_without_properties(self) -> None:
        defined_classes = {class_.class_ for class_ in self._classes}
        referred_classes = {property_.class_ for property_ in self._properties}
        class_parent_pairs = self._class_parent_pairs()

        if classes_without_properties := defined_classes.difference(referred_classes):
            for class_ in classes_without_properties:
                # USE CASE: class has no direct properties and no parents with properties
                # and it is a class in the prefix of data model, as long as it is in the
                # same prefix, meaning same space
                if not class_parent_pairs[class_] and class_.prefix == self._metadata.prefix:
                    self.issue_list.append(
                        ResourceNotDefinedWarning(
                            resource_type="class",
                            identifier=class_,
                            location="Properties sheet",
                        )
                    )

    def _undefined_classes(self) -> None:
        defined_classes = {class_.class_ for class_ in self._classes}
        referred_classes = {property_.class_ for property_ in self._properties}

        if undefined_classes := referred_classes.difference(defined_classes):
            for class_ in undefined_classes:
                self.issue_list.append(
                    ResourceNotDefinedError(
                        identifier=class_,
                        resource_type="class",
                        location="Classes sheet",
                    )
                )

    def _parent_class_defined(self) -> None:
        """This is a validation to check if the parent class of a class is defined in the classes sheet."""
        class_parent_pairs = self._class_parent_pairs()
        classes = set(class_parent_pairs.keys())
        parents = set(itertools.chain.from_iterable(class_parent_pairs.values()))

        if undefined_parents := parents.difference(classes):
            for parent in undefined_parents:
                if parent.prefix != self._metadata.prefix:
                    self.issue_list.append(UndefinedClassWarning(class_id=str(parent)))
                else:
                    self.issue_list.append(
                        ResourceNotDefinedWarning(
                            resource_type="class",
                            identifier=parent,
                            location="Classes sheet",
                        )
                    )

    def _referenced_classes_exist(self) -> None:
        # needs to be complete for this validation to pass
        defined_classes = {class_.class_ for class_ in self._classes}
        classes_with_explicit_properties = {property_.class_ for property_ in self._properties}

        # USE CASE: models are complete
        if missing_classes := classes_with_explicit_properties.difference(defined_classes):
            for class_ in missing_classes:
                self.issue_list.append(
                    ResourceNotDefinedWarning(
                        resource_type="class",
                        identifier=class_,
                        location="Classes sheet",
                    )
                )

    def _referenced_value_types_exist(self) -> None:
        # adding UnknownEntity to the set of defined classes to handle the case where a property references an unknown
        defined_classes = {class_.class_ for class_ in self._classes} | {UnknownEntity()}
        referred_object_types = {
            property_.value_type
            for property_ in self.rules.properties
            if property_.type_ == EntityTypes.object_property
        }

        if missing_value_types := referred_object_types.difference(defined_classes):
            # Todo: include row and column number
            for missing in missing_value_types:
                self.issue_list.append(
                    ResourceNotDefinedWarning(
                        resource_type="class",
                        identifier=missing,
                        location="Classes sheet",
                    )
                )

    def _regex_compliance_with_dms(self) -> None:
        """Check regex compliance with DMS of properties, classes and value types."""

        for prop_ in self._properties:
            if not PATTERNS.dms_property_id_compliance.match(prop_.property_):
                self.issue_list.append(
                    ResourceRegexViolationWarning(
                        prop_.property_,
                        "Property",
                        "Properties sheet, Property column",
                        PATTERNS.dms_property_id_compliance.pattern,
                    )
                )
            if not PATTERNS.view_id_compliance.match(prop_.class_.suffix):
                self.issue_list.append(
                    ResourceRegexViolationWarning(
                        prop_.class_,
                        "Class",
                        "Properties sheet, Class column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )

            # Handling Value Type
            if (
                isinstance(prop_.value_type, ClassEntity)
                and prop_.value_type != UnknownEntity()
                and not PATTERNS.view_id_compliance.match(prop_.value_type.suffix)
            ):
                self.issue_list.append(
                    ResourceRegexViolationWarning(
                        prop_.value_type,
                        "Value Type",
                        "Properties sheet, Value Type column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )
            if isinstance(prop_.value_type, MultiValueTypeInfo):
                for value_type in prop_.value_type.types:
                    if (
                        isinstance(prop_.value_type, ClassEntity)
                        and prop_.value_type != UnknownEntity()
                        and not PATTERNS.view_id_compliance.match(value_type.suffix)
                    ):
                        self.issue_list.append(
                            ResourceRegexViolationWarning(
                                value_type,
                                "Value Type",
                                "Properties sheet, Value Type column",
                                PATTERNS.view_id_compliance.pattern,
                            )
                        )

        for class_ in self._classes:
            if not PATTERNS.view_id_compliance.match(class_.class_.suffix):
                self.issue_list.append(
                    ResourceRegexViolationWarning(
                        class_.class_,
                        "Class",
                        "Classes sheet, Class column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )

            if class_.implements:
                for parent in class_.implements:
                    if not PATTERNS.view_id_compliance.match(parent.suffix):
                        self.issue_list.append(
                            ResourceRegexViolationWarning(
                                parent,
                                "Class",
                                "Classes sheet, Implements column",
                                PATTERNS.view_id_compliance.pattern,
                            )
                        )

    def _class_parent_pairs(self) -> dict[ClassEntity, list[ClassEntity]]:
        class_parent_pairs: dict[ClassEntity, list[ClassEntity]] = {}
        classes = self.rules.model_copy(deep=True).classes

        for class_ in classes:
            class_parent_pairs[class_.class_] = []
            if class_.implements is None:
                continue
            class_parent_pairs[class_.class_].extend(class_.implements)

        return class_parent_pairs

    def _namespaces_reassigned(self) -> None:
        prefixes = self.rules.prefixes.copy()
        prefixes[self.rules.metadata.namespace.prefix] = self.rules.metadata.namespace

        if len(set(prefixes.values())) != len(prefixes):
            reused_namespaces = [value for value, count in Counter(prefixes.values()).items() if count > 1]
            impacted_prefixes = [key for key, value in prefixes.items() if value in reused_namespaces]
            self.issue_list.append(
                NeatValueError(
                    "Namespace collision detected. The following prefixes "
                    f"are assigned to the same namespace: {impacted_prefixes}"
                    f"\nImpacted namespaces: {reused_namespaces}"
                    "\nMake sure that each unique namespace is assigned to a unique prefix"
                )
            )



import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Collection, Iterable, Sequence
from dataclasses import dataclass, field
from graphlib import TopologicalSorter
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, TypeVar, cast, overload

from cognite.client.data_classes import filters
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelList,
    EdgeConnection,
    MappedProperty,
    Node,
    NodeApply,
    NodeApplyList,
    NodeList,
    RequiresConstraint,
    Space,
    SpaceApply,
    SpaceApplyList,
    SpaceList,
    View,
    ViewApply,
    ViewApplyList,
    ViewList,
)
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    DataModelId,
    NodeId,
    ViewId,
)
from cognite.client.data_classes.data_modeling.views import (
    EdgeConnectionApply,
    MappedPropertyApply,
    ReverseDirectRelation,
    ReverseDirectRelationApply,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.neat._client.data_classes.data_modeling import Component
from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._issues.warnings import CDFMaxIterationsWarning
from cognite.neat._shared import T_ID

if TYPE_CHECKING:
    from cognite.neat._client._api_client import NeatClient

T_WritableCogniteResourceList = TypeVar("T_WritableCogniteResourceList", bound=WriteableCogniteResourceList)

T_Item = TypeVar("T_Item")
T_Out = TypeVar("T_Out", bound=Iterable)


@dataclass
class MultiCogniteAPIError(Exception, Generic[T_ID, T_WritableCogniteResourceList]):
    success: T_WritableCogniteResourceList
    failed: list[T_ID] = field(default_factory=list)
    errors: list[CogniteAPIError] = field(default_factory=list)


class ResourceLoader(
    ABC,
    Generic[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
):
    """A resource loaders is a wrapper around the Cognite SDK that does a few things:

    * Standardizes the CRUD operations for a given resource.
    * Caches the items that have been retrieved from the CDF.
    """

    resource_name: str
    dependencies: "ClassVar[frozenset[type[ResourceLoader]]]" = frozenset()

    def __init__(self, client: "NeatClient") -> None:
        # This is exposed to allow for disabling the cache.
        self.cache = True
        self._client = client
        # This cache is used to store the items that have been retrieved from the CDF.
        self._items_by_id: dict[T_ID, T_WritableCogniteResource] = {}

    def bust_cache(self) -> None:
        self._items_by_id = {}

    @classmethod
    @abstractmethod
    def get_id(cls, item: T_WriteClass | T_WritableCogniteResource | dict | T_ID) -> T_ID:
        raise NotImplementedError

    @classmethod
    def get_ids(cls, items: Sequence[T_WriteClass | T_WritableCogniteResource]) -> list[T_ID]:
        return [cls.get_id(item) for item in items]

    def create(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        # Containers can have dependencies on other containers, so we sort them before creating them.
        items = self.sort_by_dependencies(items)

        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        try:
            created = self._fallback_one_by_one(self._create, items)
        except MultiCogniteAPIError as e:
            created = e.success
            exception = e

        if self.cache:
            self._items_by_id.update({self.get_id(item): item for item in created})

        if exception is not None:
            raise exception

        return created

    @overload
    def retrieve(
        self, ids: SequenceNotStr[T_ID], format: Literal["read"] = "read"
    ) -> T_WritableCogniteResourceList: ...

    @overload
    def retrieve(self, ids: SequenceNotStr[T_ID], format: Literal["write"] = "write") -> T_CogniteResourceList: ...

    def retrieve(
        self, ids: SequenceNotStr[T_ID], format: Literal["read", "write"] = "read"
    ) -> T_WritableCogniteResourceList | T_CogniteResourceList:
        if not self.cache:
            # We now that SequenceNotStr = Sequence
            output = self._fallback_one_by_one(self._retrieve, ids)  # type: ignore[arg-type]
        else:
            exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
            missing_ids = [id for id in ids if id not in self._items_by_id.keys()]
            if missing_ids:
                try:
                    retrieved = self._retrieve(missing_ids)
                except MultiCogniteAPIError as e:
                    retrieved = e.success
                    exception = e
                self._items_by_id.update({self.get_id(item): item for item in retrieved})
            if exception is not None:
                raise exception
            # We need to check the cache again, in case we didn't retrieve all the items.
            output = self._create_list([self._items_by_id[id] for id in ids if id in self._items_by_id])
        if format == "write":
            return cast(T_CogniteResourceList, output.as_write())
        return output

    def update(
        self, items: Sequence[T_WriteClass], force: bool = False, drop_data: bool = False
    ) -> T_WritableCogniteResourceList:
        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        if force:
            updated = self._update_force(items, drop_data=drop_data)
        else:
            try:
                updated = self._fallback_one_by_one(self._update, items)
            except MultiCogniteAPIError as e:
                updated = e.success
                exception = e

        if self.cache:
            self._items_by_id.update({self.get_id(item): item for item in updated})

        if exception is not None:
            raise exception

        return updated

    def delete(self, ids: SequenceNotStr[T_ID] | Sequence[T_WriteClass]) -> list[T_ID]:
        id_list = [self.get_id(item) for item in ids]
        exception: MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList] | None = None
        try:
            # We know that SequenceNotStr = Sequence
            deleted = self._fallback_one_by_one(self._delete, id_list)  # type: ignore[arg-type]
        except MultiCogniteAPIError as e:
            deleted = e.success
            exception = e

        if self.cache:
            for id in deleted:
                self._items_by_id.pop(id, None)
        if exception is not None:
            raise exception

        return deleted

    @abstractmethod
    def _create(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _retrieve(self, ids: SequenceNotStr[T_ID]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _update(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def _delete(self, ids: SequenceNotStr[T_ID]) -> list[T_ID]:
        raise NotImplementedError

    @abstractmethod
    def _create_list(self, items: Sequence[T_WritableCogniteResource]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    def has_data(self, item_id: T_ID) -> bool:
        return False

    def are_equal(self, local: T_WriteClass, remote: T_WritableCogniteResource) -> bool:
        return local == remote.as_write()

    def sort_by_dependencies(self, items: Sequence[T_WriteClass]) -> list[T_WriteClass]:
        return list(items)

    def _update_force(
        self,
        items: Sequence[T_WriteClass],
        drop_data: bool = False,
        tried_force_update: set[T_ID] | None = None,
        success: T_WritableCogniteResourceList | None = None,
    ) -> T_WritableCogniteResourceList:
        tried_force_update = tried_force_update or set()
        try:
            return self._update(items)
        except CogniteAPIError as e:
            failed_ids = {self.get_id(failed) for failed in e.failed + e.unknown}
            success_ids = [self.get_id(success) for success in e.successful]
            success_ = self.retrieve(success_ids)
            if success is None:
                success = success_
            else:
                success.extend(success_)
            to_redeploy: list[T_WriteClass] = []
            for item in items:
                item_id = self.get_id(item)
                if item_id in failed_ids:
                    if tried_force_update and item_id in tried_force_update:
                        # Avoid infinite loop
                        continue
                    tried_force_update.add(item_id)
                    if self.has_data(item_id) and not drop_data:
                        continue
                    to_redeploy.append(item)
            if not to_redeploy:
                # Avoid infinite loop
                raise e
            self.delete(to_redeploy)
            forced = self._update_force(to_redeploy, drop_data, tried_force_update, success)
            forced.extend(success)
            return forced

    def _fallback_one_by_one(self, method: Callable[[Sequence[T_Item]], T_Out], items: Sequence[T_Item]) -> T_Out:
        try:
            return method(items)
        except CogniteAPIError as e:
            exception = MultiCogniteAPIError[T_ID, T_WritableCogniteResourceList](self._create_list([]))
            success = {self.get_id(success) for success in e.successful}
            if success:
                # Need read version of the items to put into cache.
                retrieve_items = self.retrieve(list(success))
                exception.success.extend(retrieve_items)
            for item in items:
                # We know that item is either T_ID or T_WriteClass
                # but the T_Item cannot be bound to both types at the same time.
                item_id = self.get_id(item)  # type: ignore[arg-type]
                if item_id in success:
                    continue
                try:
                    item_result = method([item])
                except CogniteAPIError as item_exception:
                    exception.errors.append(item_exception)
                    exception.failed.extend(self.get_ids(item_exception.failed))
                else:
                    exception.success.extend(item_result)
            raise exception from None


class DataModelingLoader(
    ResourceLoader[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
    ABC,
):
    @classmethod
    def in_space(cls, item: T_WriteClass | T_WritableCogniteResource | T_ID, space: set[str]) -> bool:
        if hasattr(item, "space"):
            return item.space in space
        raise ValueError(f"Item {item} does not have a space attribute")

    @classmethod
    @abstractmethod
    def items_from_schema(cls, schema: DMSSchema) -> T_CogniteResourceList:
        raise NotImplementedError


class SpaceLoader(DataModelingLoader[str, SpaceApply, Space, SpaceApplyList, SpaceList]):
    resource_name = "spaces"

    @classmethod
    def get_id(cls, item: Space | SpaceApply | str | dict) -> str:
        if isinstance(item, Space | SpaceApply):
            return item.space
        if isinstance(item, dict):
            return item["space"]
        return item

    def _create(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self._client.data_modeling.spaces.apply(items)

    def _retrieve(self, ids: SequenceNotStr[str]) -> SpaceList:
        return self._client.data_modeling.spaces.retrieve(ids)

    def _update(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[str] | Sequence[Space | SpaceApply]) -> list[str]:
        if all(isinstance(item, Space) for item in ids) or all(isinstance(item, SpaceApply) for item in ids):
            ids = [cast(Space | SpaceApply, item).space for item in ids]
        return self._client.data_modeling.spaces.delete(cast(SequenceNotStr[str], ids))

    def _create_list(self, items: Sequence[Space]) -> SpaceList:
        return SpaceList(items)

    def clean(self, space: str) -> None:
        """Deletes all data in a space.

        This means all nodes, edges, views, containers, and data models located in the given space.

        Args:
            client: Connected CogniteClient
            space: The space to delete.

        """
        edges = self._client.data_modeling.instances.list(
            "edge", limit=-1, filter=filters.Equals(["edge", "space"], space)
        )
        if edges:
            instances = self._client.data_modeling.instances.delete(edges=edges.as_ids())
            print(f"Deleted {len(instances.edges)} edges")
        nodes = self._client.data_modeling.instances.list(
            "node", limit=-1, filter=filters.Equals(["node", "space"], space)
        )
        node_types = {NodeId(node.type.space, node.type.external_id) for node in nodes if node.type}
        node_data = set(nodes.as_ids()) - node_types
        if node_data:
            instances = self._client.data_modeling.instances.delete(nodes=list(node_data))
            print(f"Deleted {len(instances.nodes)} nodes")
        if node_types:
            instances = self._client.data_modeling.instances.delete(nodes=list(node_types))
            print(f"Deleted {len(instances.nodes)} node types")
        views = self._client.data_modeling.views.list(limit=-1, space=space)
        if views:
            deleted_views = self._client.data_modeling.views.delete(views.as_ids())
            print(f"Deleted {len(deleted_views)} views")
        containers = self._client.data_modeling.containers.list(limit=-1, space=space)
        if containers:
            deleted_containers = self._client.data_modeling.containers.delete(containers.as_ids())
            print(f"Deleted {len(deleted_containers)} containers")
        if data_models := self._client.data_modeling.data_models.list(limit=-1, space=space):
            deleted_data_models = self._client.data_modeling.data_models.delete(data_models.as_ids())
            print(f"Deleted {len(deleted_data_models)} data models")
        deleted_space = self._client.data_modeling.spaces.delete(space)
        print(f"Deleted space {deleted_space}")

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> SpaceApplyList:
        return SpaceApplyList(schema.spaces.values())

    def has_data(self, item_id: str) -> bool:
        if self._client.data_modeling.instances.list("node", limit=1, space=item_id):
            return True
        if self._client.data_modeling.instances.list("edge", limit=1, space=item_id):
            return True
        # Need to check if there are any containers with data in the space. Typically,
        # a schema space will not contain data, while it will have containers that have data in an instance space.
        for container in self._client.data_modeling.containers(space=item_id, include_global=False):
            if self._client.loaders.containers.has_data(container.as_id()):
                return True
        return False


class ContainerLoader(DataModelingLoader[ContainerId, ContainerApply, Container, ContainerApplyList, ContainerList]):
    resource_name = "containers"
    dependencies = frozenset({SpaceLoader})

    @classmethod
    def get_id(cls, item: Container | ContainerApply | ContainerId | dict) -> ContainerId:
        if isinstance(item, Container | ContainerApply):
            return item.as_id()
        if isinstance(item, dict):
            return ContainerId.load(item)
        return item

    def sort_by_dependencies(self, items: Sequence[ContainerApply]) -> list[ContainerApply]:
        container_by_id = {container.as_id(): container for container in items}
        container_dependencies = {
            container.as_id(): {
                const.require
                for const in container.constraints.values()
                if isinstance(const, RequiresConstraint) and const.require in container_by_id
            }
            for container in items
        }
        return [
            container_by_id[container_id] for container_id in TopologicalSorter(container_dependencies).static_order()
        ]

    def _create(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self._client.data_modeling.containers.apply(items)

    @overload
    def retrieve(
        self, ids: SequenceNotStr[ContainerId], format: Literal["read"] = "read", include_connected: bool = False
    ) -> ContainerList: ...

    @overload
    def retrieve(
        self, ids: SequenceNotStr[ContainerId], format: Literal["write"] = "write", include_connected: bool = False
    ) -> ContainerApplyList: ...

    def retrieve(
        self,
        ids: SequenceNotStr[ContainerId],
        format: Literal["read", "write"] = "read",
        include_connected: bool = False,
    ) -> ContainerList | ContainerApplyList:
        if not include_connected:
            return super().retrieve(ids)

        # Retrieve recursively updates the cache.
        output = self._retrieve_recursive(ids)
        if format == "write":
            return output.as_write()
        return output

    def _retrieve(self, ids: SequenceNotStr[ContainerId]) -> ContainerList:
        return self._client.data_modeling.containers.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[ContainerId]) -> list[ContainerId]:
        return self._client.data_modeling.containers.delete(cast(Sequence, ids))

    def _create_list(self, items: Sequence[Container]) -> ContainerList:
        return ContainerList(items)

    def _retrieve_recursive(self, container_ids: SequenceNotStr[ContainerId]) -> ContainerList:
        """Containers can reference each other through the 'requires' constraint.

        This method retrieves all containers that are referenced by other containers through the 'requires' constraint,
        including their parents.
        """
        max_iterations = 10  # Limiting the number of iterations to avoid infinite loops
        found = ContainerList([])
        found_ids: set[ContainerId] = set()
        last_batch = list(container_ids)
        for _ in range(max_iterations):
            if not last_batch:
                break
            to_retrieve_from_cdf: set[ContainerId] = set()
            batch_ids: list[ContainerId] = []
            for container_id in last_batch:
                if container_id in found_ids:
                    continue
                elif container_id in self._items_by_id:
                    container = self._items_by_id[container_id]
                    found.append(container)
                    batch_ids.extend(self.get_connected_containers(container, found_ids))
                else:
                    to_retrieve_from_cdf.add(container_id)

            if to_retrieve_from_cdf:
                retrieved_batch = self._client.data_modeling.containers.retrieve(list(to_retrieve_from_cdf))
                self._items_by_id.update({view.as_id(): view for view in retrieved_batch})
                found.extend(retrieved_batch)
                found_ids.update({view.as_id() for view in retrieved_batch})
                for container in retrieved_batch:
                    batch_ids.extend(self.get_connected_containers(container, found_ids))

            last_batch = batch_ids
        else:
            warnings.warn(
                CDFMaxIterationsWarning(
                    "The maximum number of iterations was reached while resolving referenced containers."
                    "There might be referenced containers that are not included in the list of containers.",
                    max_iterations=max_iterations,
                ),
                stacklevel=2,
            )

        if self.cache is False:
            # We must update the cache to retrieve recursively.
            # If the cache is disabled, bust the cache to avoid storing the retrieved views.
            self.bust_cache()
        return found

    @staticmethod
    def get_connected_containers(
        container: Container | ContainerApply, skip: set[ContainerId] | None = None
    ) -> set[ContainerId]:
        connected_containers = set()
        for constraint in container.constraints.values():
            if isinstance(constraint, RequiresConstraint):
                connected_containers.add(constraint.require)
        if skip:
            return {container_id for container_id in connected_containers if container_id not in skip}
        return connected_containers

    def are_equal(self, local: ContainerApply, remote: Container) -> bool:
        local_dumped = local.dump(camel_case=True)
        if "usedFor" not in local_dumped:
            # Setting used_for to "node" as it is the default value in the CDF.
            local_dumped["usedFor"] = "node"

        return local_dumped == remote.as_write().dump(camel_case=True)

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> ContainerApplyList:
        return ContainerApplyList(schema.containers.values())

    def has_data(self, item_id: ContainerId) -> bool:
        has_data_filter = filters.HasData(containers=[item_id])
        has_data = False
        instance_type: Literal["node", "edge"]
        # Mypy does not understand that the instance type is Literal["node", "edge"]
        for instance_type in ["node", "edge"]:  # type: ignore[assignment]
            try:
                has_data = bool(
                    self._client.data_modeling.instances.list(instance_type, limit=1, filter=has_data_filter)
                )
            except CogniteAPIError as e:
                if e.code != 400:
                    # If the container is used for nodes and we ask for edges, we get a 400 error. This
                    # means there is no edge data for this container.
                    raise
            if has_data:
                return True
        return has_data


class ViewLoader(DataModelingLoader[ViewId, ViewApply, View, ViewApplyList, ViewList]):
    resource_name = "views"
    dependencies = frozenset({SpaceLoader, ContainerLoader})

    @classmethod
    def get_id(cls, item: View | ViewApply | ViewId | dict) -> ViewId:
        if isinstance(item, View | ViewApply):
            return item.as_id()
        if isinstance(item, dict):
            return ViewId.load(item)
        return item

    def _create(self, items: Sequence[ViewApply]) -> ViewList:
        return self._client.data_modeling.views.apply(items)

    @overload
    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["read"] = "read",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewList: ...

    @overload
    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["write"] = "write",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewApplyList: ...

    def retrieve(
        self,
        ids: SequenceNotStr[ViewId],
        format: Literal["read", "write"] = "read",
        include_connected: bool = False,
        include_ancestor: bool = False,
    ) -> ViewList | ViewApplyList:
        if not include_connected and not include_ancestor:
            # Default .as_write() method does not work for views as they include parent properties.
            output = super().retrieve(ids)
        else:
            # Retrieve recursively updates the cache.
            output = self._retrieve_recursive(ids, include_connected, include_ancestor)
        if format == "write":
            return ViewApplyList([self.as_write(view) for view in output])
        return output

    def _retrieve(self, ids: SequenceNotStr[ViewId]) -> ViewList:
        return self._client.data_modeling.views.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[ViewApply]) -> ViewList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[ViewId]) -> list[ViewId]:
        return self._client.data_modeling.views.delete(cast(Sequence, ids))

    def _as_write_raw(self, view: View) -> dict[str, Any]:
        dumped = view.as_write().dump()
        if view.properties:
            # All read version of views have all the properties of their parent views.
            # We need to remove these properties to compare with the local view.
            parents = self._retrieve_recursive(view.implements or [], include_connections=False, include_ancestors=True)
            for parent in parents:
                for prop_name, prop in (parent.as_write().properties or {}).items():
                    existing = dumped["properties"].get(prop_name)
                    if existing is None:
                        continue
                    if existing == prop.dump():
                        dumped["properties"].pop(prop_name, None)
                    # If the child overrides the parent, we keep the child's property.

        if "properties" in dumped and not dumped["properties"]:
            # All properties were removed, so we remove the properties key.
            dumped.pop("properties", None)
        return dumped

    def are_equal(self, local: ViewApply, remote: View) -> bool:
        local_dumped = local.dump()
        if not remote.implements:
            return local_dumped == remote.as_write().dump()

        cdf_resource_dumped = self._as_write_raw(remote)

        if "properties" in local_dumped and not local_dumped["properties"]:
            # In case the local properties are set to an empty dict.
            local_dumped.pop("properties", None)

        return local_dumped == cdf_resource_dumped

    def as_write(self, view: View) -> ViewApply:
        return ViewApply.load(self._as_write_raw(view))

    def _retrieve_recursive(
        self, view_ids: SequenceNotStr[ViewId], include_connections: bool, include_ancestors: bool
    ) -> ViewList:
        """Retrieves all views with the

        This will mutate the cache passed in, and return a list of views that are the ancestors
        of the views in the parents list.

        Args:
            view_ids: The views to retrieve.
            include_connections: Whether to include all connected views.
            include_ancestors: Whether to include all ancestors.
        """
        last_batch = set(view_ids)
        found = ViewList([])
        found_ids: set[ViewId] = set()
        while last_batch:
            to_retrieve_from_cdf: set[ViewId] = set()
            batch_ids: set[ViewId] = set()
            for view_id in last_batch:
                if view_id in found_ids:
                    continue
                elif view_id in self._items_by_id:
                    view = self._items_by_id[view_id]
                    found.append(view)
                    found_ids.add(view_id)
                    batch_ids.update(self.get_connected_views(view, include_ancestors, include_connections, found_ids))
                else:
                    to_retrieve_from_cdf.add(view_id)

            if to_retrieve_from_cdf:
                retrieved_batch = self._client.data_modeling.views.retrieve(list(to_retrieve_from_cdf))
                self._items_by_id.update({view.as_id(): view for view in retrieved_batch})
                found.extend(retrieved_batch)
                found_ids.update({view.as_id() for view in retrieved_batch})
                for view in retrieved_batch:
                    batch_ids.update(self.get_connected_views(view, include_ancestors, include_connections, found_ids))
            last_batch = batch_ids

        if self.cache is False:
            # We must update the cache to retrieve recursively.
            # If the cache is disabled, bust the cache to avoid storing the retrieved views.
            self.bust_cache()
        return found

    @staticmethod
    def get_connected_views(
        view: View | ViewApply,
        include_parents: bool = True,
        include_connections: bool = True,
        skip_ids: set[ViewId] | None = None,
    ) -> list[ViewId]:
        connected_ids: set[ViewId] = set(view.implements or []) if include_parents else set()
        if include_connections:
            for prop in (view.properties or {}).values():
                if isinstance(prop, MappedProperty | MappedPropertyApply) and prop.source:
                    connected_ids.add(prop.source)
                elif isinstance(
                    prop, EdgeConnection | EdgeConnectionApply | ReverseDirectRelation | ReverseDirectRelationApply
                ):
                    connected_ids.add(prop.source)

                if isinstance(prop, EdgeConnection | EdgeConnectionApply) and prop.edge_source:
                    connected_ids.add(prop.edge_source)
                elif isinstance(prop, ReverseDirectRelation | ReverseDirectRelationApply) and isinstance(
                    prop.through.source, ViewId
                ):
                    connected_ids.add(prop.through.source)
        if skip_ids:
            return [view_id for view_id in connected_ids if view_id not in skip_ids]
        return list(connected_ids)

    def _create_list(self, items: Sequence[View]) -> ViewList:
        return ViewList(items)

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> ViewApplyList:
        return ViewApplyList(schema.views.values())


class DataModelLoader(DataModelingLoader[DataModelId, DataModelApply, DataModel, DataModelApplyList, DataModelList]):
    resource_name = "data_models"
    dependencies = frozenset({SpaceLoader, ViewLoader})

    @classmethod
    def get_id(cls, item: DataModel | DataModelApply | DataModelId | dict) -> DataModelId:
        if isinstance(item, DataModel | DataModelApply):
            return item.as_id()
        if isinstance(item, dict):
            return DataModelId.load(item)
        return item

    def _create(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self._client.data_modeling.data_models.apply(items)

    def _retrieve(self, ids: SequenceNotStr[DataModelId]) -> DataModelList:
        return self._client.data_modeling.data_models.retrieve(cast(Sequence, ids))

    def _update(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self._create(items)

    def _delete(self, ids: SequenceNotStr[DataModelId]) -> list[DataModelId]:
        return self._client.data_modeling.data_models.delete(cast(Sequence, ids))

    def _create_list(self, items: Sequence[DataModel]) -> DataModelList:
        return DataModelList(items)

    def are_equal(self, local: DataModelApply, remote: DataModel) -> bool:
        local_dumped = local.dump()
        cdf_resource_dumped = remote.as_write().dump()

        # Data models that have the same views, but in different order, are considered equal.
        # We also account for whether views are given as IDs or View objects.
        local_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in local.views or []
        )
        cdf_resource_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in remote.views or []
        )

        return local_dumped == cdf_resource_dumped

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> DataModelApplyList:
        return DataModelApplyList([schema.data_model])


class NodeLoader(DataModelingLoader[NodeId, NodeApply, Node, NodeApplyList, NodeList]):
    resource_name = "nodes"
    dependencies = frozenset({SpaceLoader, ContainerLoader, ViewLoader})

    @classmethod
    def get_id(cls, item: Node | NodeApply | NodeId | dict) -> NodeId:
        if isinstance(item, Node | NodeApply):
            return item.as_id()
        if isinstance(item, dict):
            return NodeId.load(item)
        return item

    def _create(self, items: Sequence[NodeApply]) -> NodeList:
        self._client.data_modeling.instances.apply(items)
        return self._retrieve([item.as_id() for item in items])

    def _retrieve(self, ids: SequenceNotStr[NodeId]) -> NodeList:
        return self._client.data_modeling.instances.retrieve(cast(Sequence, ids)).nodes

    def _update(self, items: Sequence[NodeApply]) -> NodeList:
        self._client.data_modeling.instances.apply(items, replace=True)
        return self._retrieve([item.as_id() for item in items])

    def _delete(self, ids: SequenceNotStr[NodeId]) -> list[NodeId]:
        return list(self._client.data_modeling.instances.delete(nodes=cast(Sequence, ids)).nodes)

    def _create_list(self, items: Sequence[Node]) -> NodeList:
        return NodeList(items)

    def are_equal(self, local: NodeApply, remote: Node) -> bool:
        local_dumped = local.dump()

        # Note reading from a container is not supported.
        sources = [
            source_prop_pair.source
            for source_prop_pair in local.sources or []
            if isinstance(source_prop_pair.source, ViewId)
        ]
        if sources:
            try:
                cdf_resource_with_properties = self._client.data_modeling.instances.retrieve(
                    nodes=remote.as_id(), sources=sources
                ).nodes[0]
            except CogniteAPIError:
                # View does not exist, so node does not exist.
                return False
        else:
            cdf_resource_with_properties = remote
        cdf_resource_dumped = cdf_resource_with_properties.as_write().dump()

        if "existingVersion" not in local_dumped:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            local_dumped["existingVersion"] = cdf_resource_dumped.get("existingVersion", None)

        return local_dumped == cdf_resource_dumped

    @classmethod
    def items_from_schema(cls, schema: DMSSchema) -> NodeApplyList:
        return NodeApplyList(schema.node_types.values())


class DataModelLoaderAPI:
    def __init__(self, client: "NeatClient") -> None:
        self._client = client
        self.spaces = SpaceLoader(client)
        self.views = ViewLoader(client)
        self.containers = ContainerLoader(client)
        self.data_models = DataModelLoader(client)
        self.nodes = NodeLoader(client)
        self._loaders: list[DataModelingLoader] = [
            self.spaces,
            self.views,
            self.containers,
            self.data_models,
            self.nodes,
        ]

    def by_dependency_order(
        self, component: Component | Collection[Component] | None = None
    ) -> list[DataModelingLoader]:
        loader_by_type = {type(loader): loader for loader in self._loaders}
        loader_iterable = (
            loader_by_type[loader_cls]  # type: ignore[index]
            for loader_cls in TopologicalSorter(
                {type(loader): loader.dependencies for loader in self._loaders}  # type: ignore[attr-defined]
            ).static_order()
        )
        if component is None:
            return list(loader_iterable)
        components = {component} if isinstance(component, str) else set(component)
        components = {{"node_type": "nodes"}.get(component, component) for component in components}
        return [loader for loader in loader_iterable if loader.resource_name in components]



from cognite.neat._issues._base import IssueList
from cognite.neat._rules.transformers import (
    ToCompliantEntities,
)

from ._state import SessionState
from .exceptions import session_class_wrapper


@session_class_wrapper
class FixAPI:
    """Apply variety of fix methods to data model and instances"""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.data_model = DataModelFixAPI(state, verbose)


@session_class_wrapper
class DataModelFixAPI:
    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def cdf_compliant_external_ids(self) -> IssueList:
        """Convert (information/logical) data model component external ids to CDF compliant entities."""
        return self._state.rule_transform(ToCompliantEntities())



from cognite.neat._issues._base import NeatError, _get_subclasses

from ._external import (
    AuthorizationError,
    CDFMissingClientError,
    CDFMissingResourcesError,
    FileMissingRequiredFieldError,
    FileNotAFileError,
    FileNotFoundNeatError,
    FileReadError,
    FileTypeUnexpectedError,
    NeatYamlError,
    OxigraphStorageLockedError,
)
from ._general import NeatImportError, NeatTypeError, NeatValueError, RegexViolationError
from ._properties import (
    PropertyDefinitionDuplicatedError,
    PropertyDefinitionError,
    PropertyMappingDuplicatedError,
    PropertyNotFoundError,
    PropertyTypeNotSupportedError,
    ReversedConnectionNotFeasibleError,
)
from ._resources import (
    ResourceChangedError,
    ResourceConversionError,
    ResourceCreationError,
    ResourceDuplicatedError,
    ResourceError,
    ResourceMissingIdentifierError,
    ResourceNotDefinedError,
    ResourceNotFoundError,
    ResourceRetrievalError,
)
from ._wrapper import (
    ClassValueError,
    ContainerValueError,
    EnumValueError,
    MetadataValueError,
    NodeValueError,
    PropertyValueError,
    SpreadsheetError,
    ViewValueError,
)

__all__ = [
    "AuthorizationError",
    "CDFMissingClientError",
    "CDFMissingResourcesError",
    "ClassValueError",
    "ContainerValueError",
    "EnumValueError",
    "FileMissingRequiredFieldError",
    "FileNotAFileError",
    "FileNotFoundNeatError",
    "FileReadError",
    "FileTypeUnexpectedError",
    "MetadataValueError",
    "NeatError",
    "NeatImportError",
    "NeatTypeError",
    "NeatValueError",
    "NeatYamlError",
    "NodeValueError",
    "OxigraphStorageLockedError",
    "PropertyDefinitionDuplicatedError",
    "PropertyDefinitionError",
    "PropertyMappingDuplicatedError",
    "PropertyNotFoundError",
    "PropertyTypeNotSupportedError",
    "PropertyValueError",
    "RegexViolationError",
    "ResourceChangedError",
    "ResourceConversionError",
    "ResourceCreationError",
    "ResourceDuplicatedError",
    "ResourceError",
    "ResourceMissingIdentifierError",
    "ResourceNotDefinedError",
    "ResourceNotFoundError",
    "ResourceRetrievalError",
    "ReversedConnectionNotFeasibleError",
    "SpreadsheetError",
    "ViewValueError",
]

_NEAT_ERRORS_BY_NAME = {error.__name__: error for error in _get_subclasses(NeatError, include_base=True)}



import re
import urllib.parse
from collections.abc import Collection, Set
from re import Pattern
from typing import Any

from cognite.neat._rules._constants import get_reserved_words

PREPOSITIONS = frozenset(
    {
        "in",
        "on",
        "at",
        "by",
        "for",
        "with",
        "about",
        "against",
        "between",
        "into",
        "through",
        "during",
        "before",
        "after",
        "above",
        "below",
        "to",
        "from",
        "up",
        "down",
        "out",
        "off",
        "over",
        "under",
        "again",
        "further",
        "then",
        "once",
    }
)


def to_camel_case(string: str) -> str:
    """Convert snake_case_name to camelCaseName.

    Args:
        string: The string to convert.
    Returns:
        camelCase of the input string.

    Examples:
        >>> to_camel_case("a_b")
        'aB'
        >>> to_camel_case("ScenarioInstance_priceForecast")
        'scenarioInstancePriceForecast'
    """
    string = re.sub(r"[^a-zA-Z0-9_]", "_", string)
    string = re.sub("_+", "_", string)
    is_all_upper = string.upper() == string
    is_first_upper = (
        len(string) >= 2 and string[:2].upper() == string[:2] and "_" not in string[:2] and not is_all_upper
    )
    return _to_camel_case(string, is_all_upper, is_first_upper)


def _to_camel_case(string, is_all_upper: bool, is_first_upper: bool):
    if "_" in string:
        pascal_splits = [
            _to_pascal_case(part, is_all_upper, is_first_upper and no == 0)
            for no, part in enumerate(string.split("_"), 0)
        ]
    else:
        # Ensure pascal
        if string:
            string = string[0].upper() + string[1:]
        pascal_splits = [string]
    cleaned: list[str] = []
    for part in pascal_splits:
        if part.upper() == part and is_all_upper:
            cleaned.append(part.capitalize())
        else:
            cleaned.append(part)

    string_split = []
    for part in cleaned:
        string_split.extend(re.findall(r"[A-Z][a-z0-9]*", part))
    if not string_split:
        string_split = [string]
    if len(string_split) == 0:
        return ""
    # The first word is a single letter, keep the original case
    if is_first_upper:
        return "".join(word for word in string_split)
    else:
        return string_split[0].casefold() + "".join(word for word in string_split[1:])


def to_pascal_case(string: str) -> str:
    """Convert string to PascalCaseName.

    Args:
        string: The string to convert.
    Returns:
        PascalCase of the input string.

    Examples:
        >>> to_pascal_case("a_b")
        'AB'
        >>> to_pascal_case('camel_case')
        'CamelCase'
    """
    return _to_pascal_case(string, string == string.upper(), True)


def _to_pascal_case(string: str, is_all_upper: bool, is_first_upper: bool) -> str:
    camel = _to_camel_case(string, is_all_upper, is_first_upper)
    return f"{camel[0].upper()}{camel[1:]}" if camel else ""


def to_snake_case(string: str) -> str:
    """
    Convert input string to snake_case

    Args:
        string: The string to convert.
    Returns:
        snake_case of the input string.

    Examples:
        >>> to_snake_case("aB")
        'a_b'
        >>> to_snake_case('CamelCase')
        'camel_case'
        >>> to_snake_case('camelCamelCase')
        'camel_camel_case'
        >>> to_snake_case('Camel2Camel2Case')
        'camel_2_camel_2_case'
        >>> to_snake_case('getHTTPResponseCode')
        'get_http_response_code'
        >>> to_snake_case('get200HTTPResponseCode')
        'get_200_http_response_code'
        >>> to_snake_case('getHTTP200ResponseCode')
        'get_http_200_response_code'
        >>> to_snake_case('HTTPResponseCode')
        'http_response_code'
        >>> to_snake_case('ResponseHTTP')
        'response_http'
        >>> to_snake_case('ResponseHTTP2')
        'response_http_2'
        >>> to_snake_case('Fun?!awesome')
        'fun_awesome'
        >>> to_snake_case('Fun?!Awesome')
        'fun_awesome'
        >>> to_snake_case('10CoolDudes')
        '10_cool_dudes'
        >>> to_snake_case('20coolDudes')
        '20_cool_dudes'
    """
    pattern = re.compile(r"[A-Z]?[a-z]+|[A-Z]+(?=[A-Z][a-z]|\d|\W|$)|\d+")
    if "_" in string:
        words = [word for section in string.split("_") for word in pattern.findall(section)]
    else:
        words = pattern.findall(string)
    return "_".join(map(str.lower, words))


def to_words(string: str) -> str:
    """Converts snake_case camelCase or PascalCase to words."""
    return to_snake_case(string).replace("_", " ")


def title(text: str, skip_words: Set[str] = PREPOSITIONS) -> str:
    """Converts text to title case, skipping prepositions."""
    words = (word.lower() for word in text.split())
    titled_words = (word.capitalize() if word not in skip_words else word for word in words)
    return " ".join(titled_words)


def sentence_or_string_to_camel(string: str) -> str:
    # Could be a combination of kebab and pascal/camel case
    if " " in string:
        parts = string.split(" ")
        try:
            return parts[0].casefold() + "".join(word.capitalize() for word in parts[1:])
        except IndexError:
            return ""
    else:
        return to_camel_case(string)


def replace_non_alphanumeric_with_underscore(text: str) -> str:
    return re.sub(r"\W+", "_", text)


def humanize_collection(collection: Collection[Any], /, *, sort: bool = True) -> str:
    if not collection:
        return ""
    elif len(collection) == 1:
        return str(next(iter(collection)))

    strings = (str(item) for item in collection)
    if sort:
        sequence = sorted(strings)
    else:
        sequence = list(strings)

    return f"{', '.join(sequence[:-1])} and {sequence[-1]}"


class NamingStandardization:
    _letter_number_underscore = re.compile(r"[^a-zA-Z0-9_]+")
    _letter_number_underscore_hyphen = re.compile(r"[^a-zA-Z0-9_-]+")
    _multi_underscore_pattern = re.compile(r"_+")
    _start_letter_pattern = re.compile(r"^[a-zA-Z]")

    @classmethod
    def standardize_class_str(cls, raw: str) -> str:
        clean = cls._clean_string(raw)
        if not cls._start_letter_pattern.match(clean):
            # Underscore ensure that 'Class' it treated as a separate word
            # in the to_pascale function
            clean = f"Class_{clean}"
        return to_pascal_case(clean)

    @classmethod
    def standardize_property_str(cls, raw: str) -> str:
        clean = cls._clean_string(raw)
        if not cls._start_letter_pattern.match(clean):
            # Underscore ensure that 'property' it treated as a separate word
            # in the to_camel function
            clean = f"property_{clean}"
        return to_camel_case(clean)

    @classmethod
    def standardize_space_str(cls, raw: str) -> str:
        clean = cls._clean_string(raw, cls._letter_number_underscore_hyphen)
        if not cls._start_letter_pattern.match(clean):
            clean = f"sp_{clean}"
        if clean in set(get_reserved_words("space")):
            clean = f"my_{clean}"
        if len(clean) > 43:
            clean = clean[:43]
        if not (clean[-1].isalnum()) and len(clean) == 43:
            clean = f"{clean[:-1]}x"
        elif not clean[-1].isalnum():
            clean = f"{clean}x"
        if not clean:
            raise ValueError("Space name must contain at least one letter.")
        return to_snake_case(clean)

    @classmethod
    def _clean_string(cls, raw: str, clean_pattern: Pattern[str] = _letter_number_underscore) -> str:
        raw = urllib.parse.unquote(raw)
        raw = clean_pattern.sub("_", raw)
        return cls._multi_underscore_pattern.sub("_", raw)
